import re
from typing import Optional, Tuple, TYPE_CHECKING, Dict, Union, cast, Any
from typing_extensions import TypedDict
from collections import namedtuple
import numpy as np

from qcodes import InstrumentChannel
from .message_builder import MessageBuilder
from . import constants
from .constants import ModuleKind, SlotNr, MeasurementStatus, ChannelName, ChNr

if TYPE_CHECKING:
    from .KeysightB1500_base import KeysightB1500


_FMTResponse = namedtuple('_FMTResponse', 'value status channel type')


class MeasurementNotTaken(Exception):
    pass


def fmt_response_base_parser(raw_data_val: str) -> _FMTResponse:
    """
    Parse the response from SPA for `FMT 1,0` format  into a named tuple
    with names, value (value of the data), status (Normal or with compliance
    error such as C, T, V), channel (channel number of the output data such
    as CH1,CH2), type (current 'I' or voltage 'V'). This parser is tested
    for FMT1,0 and FMT1,1 response.

    Args:
        raw_data_val: Unparsed (raw) data for the instrument.
    """

    values_separator = ','
    data_val = []
    data_status = []
    data_channel = []
    data_datatype = []

    for str_value in raw_data_val.split(values_separator):
        status = str_value[0]
        channel_id = constants.ChannelName[str_value[1]].value

        datatype = str_value[2]
        value = float(str_value[3:])

        data_val.append(value)
        data_status.append(status)
        data_channel.append(channel_id)
        data_datatype.append(datatype)

    data = _FMTResponse(data_val, data_status, data_channel, data_datatype)
    return data


def parse_module_query_response(response: str) -> Dict[SlotNr, str]:
    """
    Extract installed module information from the given string and return the
    information as a dictionary.

    Args:
        response: Response str to `UNT? 0` query.

    Returns:
        Dictionary from slot numbers to model name strings.
    """
    pattern = r";?(?P<model>\w+),(?P<revision>\d+)"

    moduleinfo = re.findall(pattern, response)

    return {
        SlotNr(slot_nr): model
        for slot_nr, (model, rev) in enumerate(moduleinfo, start=1)
        if model != "0"
    }


# pattern to match dcv experiment
_pattern_lrn = re.compile(
    r"(?P<status_dc>\w{1,3})(?P<chnr_dc>\w),(?P<voltage_dc>\d{1,3}.\d{1,4});"
    r"(?P<status_ac>\w{1,3})(?P<chnr_ac>\w),(?P<voltage_ac>\d{1,3}.\d{1,4});"
    r"(?P<status_fc>\w{1,2})(?P<chnr_fc>\w),(?P<frequency>\d{1,6}.\d{1,4})"
)


def parse_dcv_measurement_response(response: str) -> Dict[str, Union[str,
                                                                     float]]:
    """
    Extract status, channel number, value  and accompanying metadata from
    the string and return them as a dictionary.

    Args:
        response: Response str to lrn_query For the MFCMU.
    """

    match = re.match(_pattern_lrn, response)
    if match is None:
        raise ValueError(f"{response!r} didn't match {_pattern_lrn!r} pattern")

    dd = match.groupdict()
    d = cast(Dict[str, Union[str, float]], dd)
    return d


# Pattern to match the spot measurement response against
_pattern = re.compile(
    r"((?P<status>\w)(?P<channel>\w)(?P<dtype>\w))?"
    r"(?P<value>[+-]\d{1,3}\.\d{3,6}E[+-]\d{2})"
)


class SpotResponse(TypedDict):
    value: float
    status: MeasurementStatus
    channel: ChannelName
    dtype: str


def parse_spot_measurement_response(response: str) -> SpotResponse:
    """
    Extract measured value and accompanying metadata from the string
    and return them as a dictionary.

    Args:
        response: Response str to spot measurement query.

    Returns:
        Dictionary with measured value and associated metadata (e.g.
        timestamp, channel number, etc.)
    """
    match = re.match(_pattern, response)
    if match is None:
        raise ValueError(f"{response!r} didn't match {_pattern!r} pattern")

    dd = match.groupdict()

    d = SpotResponse(
        value=_convert_to_nan_if_dummy_value(float(dd["value"])),
        status=MeasurementStatus[dd["status"]],
        channel=ChannelName[dd["channel"]],
        dtype=dd["dtype"]
    )

    return d


_DCORRResponse = namedtuple('_DCORRResponse', 'mode primary secondary')


def parse_dcorr_query_response(response: str) -> _DCORRResponse:
    """
    Parse string response of ``DCORR?`` `command into a named tuple of
    :class:`constants.DCORR.Mode` and primary and secondary reference or
    calibration values.
    """
    mode, primary, secondary = response.split(',')
    return _DCORRResponse(mode=constants.DCORR.Mode(int(mode)),
                          primary=float(primary),
                          secondary=float(secondary))


def fixed_negative_float(response: str) -> float:
    """
    Keysight sometimes responds for ex. '-0.-1' as an output when you input
    '-0.1'. This function can convert such strings also to float.
    """
    if len(response.split('.')) > 2:
        raise ValueError('String must of format `a` or `a.b`')

    parts = response.split('.')
    number = parts[0]
    decimal = parts[1] if len(parts) > 1 else '0'

    decimal = decimal.replace("-", "")

    output = ".".join([number, decimal])
    return float(output)


_dcorr_labels_units_map = {
    constants.DCORR.Mode.Cp_G: dict(
        primary=dict(label='Cp', unit='F'),
        secondary=dict(label='G', unit='S')
    ),
    constants.DCORR.Mode.Ls_Rs: dict(
        primary=dict(label='Ls', unit='H'),
        secondary=dict(label='Rs', unit='Ω'))
}


def format_dcorr_response(r: _DCORRResponse) -> str:
    """
    Format a given response tuple ``_DCORRResponse`` from
    ``DCORR?`` command as a human-readable string.
    """
    labels_units = _dcorr_labels_units_map[r.mode]
    primary = labels_units['primary']
    secondary = labels_units['secondary']

    result_str = \
        f"Mode: {r.mode.name}, " \
        f"Primary {primary['label']}: {r.primary} {primary['unit']}, " \
        f"Secondary {secondary['label']}: {r.secondary} {secondary['unit']}"

    return result_str


def get_name_label_unit_of_impedance_model(
        mode: constants.IMP.MeasurementMode
) -> Tuple[Tuple[str, str], Tuple[str, str], Tuple[str, str]]:
    params = mode.name.split('_')

    param1 = params[0]
    param2 = '_'.join(params[1:])

    label = (constants.IMP.Name[param1].value,
             constants.IMP.Name[param2].value)

    unit = (constants.IMP.Unit[param1].value,
            constants.IMP.Unit[param2].value)

    name = (label[0].lower().replace(' ', '_'),
            label[1].lower().replace(' ', '_'))

    return name, label, unit


# TODO notes:
# - [ ] Instead of generating a Qcodes InstrumentChannel for each **module**,
#   it might make more sense to generate one for each **channel**


def get_measurement_summary(status_array: np.ndarray) -> str:
    unique_error_statuses = np.unique(status_array[status_array != "N"])
    if len(unique_error_statuses) > 0:
        summary = " ".join(
            constants.MeasurementStatus[err] for err in
            unique_error_statuses
        )
    else:
        summary = constants.MeasurementStatus["N"]

    return summary


def convert_dummy_val_to_nan(param: _FMTResponse) -> None:
    """
    Converts dummy value to NaN. Instrument may output dummy value (
    199.999E+99) if measurement data is over the measurement range. Or the
    sweep measurement was aborted by the automatic stop function or power
    compliance. Or if any abort condition is detected. Dummy data
    199.999E+99 will be returned for the data after abort."

    Args:
        param: This must be of type named tuple _FMTResponse.

    """
    for index, value in enumerate(param.value):
        param.value[index] = _convert_to_nan_if_dummy_value(param.value[index])


def _convert_to_nan_if_dummy_value(value: float) -> float:
    return float('nan') if value > 1e99 else value


class B1500Module(InstrumentChannel):
    """Base class for all modules of B1500 Parameter Analyzer

    When subclassing,

      - set ``MODULE_KIND`` attribute to the correct module kind
        :class:`~.constants.ModuleKind` that the module is.
      - populate ``channels`` attribute according to the number of
        channels that the module has.

    Args:
        parent: Mainframe B1500 instance that this module belongs to
        name: Name of the instrument instance to create. If `None`
            (Default), then the name is autogenerated from the instrument
            class.
        slot_nr: Slot number of this module (not channel number)
    """
    MODULE_KIND: ModuleKind

    def __init__(self, parent: 'KeysightB1500', name: Optional[str],
                 slot_nr: int,
                 **kwargs: Any):
        # self.channels will be populated in the concrete module subclasses
        # because channel count is module specific
        self.channels: Tuple[ChNr, ...]
        self.slot_nr = SlotNr(slot_nr)

        if name is None:
            number = len(parent.by_kind[self.MODULE_KIND]) + 1
            name = self.MODULE_KIND.lower() + str(number)

        super().__init__(parent=parent, name=name, **kwargs)

    # Response parsing functions as static methods for user convenience
    parse_spot_measurement_response = parse_spot_measurement_response
    parse_module_query_response = parse_module_query_response

    def enable_outputs(self) -> None:
        """
        Enables all outputs of this module by closing the output relays of its
        channels.
        """
        # TODO This always enables all outputs of a module, which is maybe not
        # desirable. (Also check the TODO item at the top about
        # InstrumentChannel per Channel instead of per Module.
        msg = MessageBuilder().cn(self.channels).message
        self.write(msg)

    def disable_outputs(self) -> None:
        """
        Disables all outputs of this module by opening the output relays of its
        channels.
        """
        # TODO See enable_output TODO item
        msg = MessageBuilder().cl(self.channels).message
        self.write(msg)

    def is_enabled(self) -> bool:
        """
        Check if channels of this module are enabled.

        Returns:
            `True` if *all* channels of this module are enabled. `False`,
            otherwise.
        """
        # TODO If a module has multiple channels, and only one is enabled, then
        # this will return false, which is probably not desirable.
        # Also check the TODO item at the top about InstrumentChannel per
        # Channel instead of per Module.
        msg = (MessageBuilder()
               .lrn_query(constants.LRN.Type.OUTPUT_SWITCH)
               .message
               )
        response = self.ask(msg)
        activated_channels = re.sub(r"[^,\d]", "", response).split(",")

        is_enabled = set(self.channels).issubset(
            int(x) for x in activated_channels if x != ''
        )
        return is_enabled

    def clear_timer_count(self) -> None:
        """
        This command clears the timer count. This command is effective for
        all measurement modes, regardless of the TSC setting. This command
        is not effective for the 4 byte binary data output format
        (FMT3 and FMT4).
        """
        self.root_instrument.clear_timer_count(chnum=self.channels)


class StatusMixin:
    def __init__(self) -> None:
        self.names = tuple(['param1', 'param2'])

    def status_summary(self) -> Dict[str, str]:
        return_dict: Dict[str, str] = {}

        for name_index, name in enumerate(self.names):
            param_data: _FMTResponse = getattr(self, f"param{name_index+1}")

            status_array = param_data.status
            if status_array is None:
                self_full_name = getattr(self, "full_name", "this")
                raise MeasurementNotTaken(
                    f"First run sweep measurement with {self_full_name} "
                    f"parameter to obtain the data; then it will be possible "
                    f"to obtain status summary for that data."
                )

            summary = get_measurement_summary(status_array)
            return_dict[name] = summary

        return return_dict



"""
Data acquisition loops.

The general scheme is:

1. create a (potentially nested) Loop, which defines the sweep setpoints and
delays

2. activate the loop (which changes it to an ActiveLoop object),

3. run it with the .run method, which creates a DataSet to hold the data,
and defines how and where to save the data.

Some examples:

- 1D sweep, using the default measurement set

>>> Loop(sweep_values, delay).run()

- 2D sweep, using the default measurement set sv1 is the outer loop, sv2 is the
  inner.

>>> Loop(sv1, delay1).loop(sv2, delay2).run()

- 1D sweep with specific measurements to take at each point

>>> Loop(sv, delay).each(param4, param5).run()

- Multidimensional sweep: 1D measurement of param6 on the outer loop, and another
  measurement in an inner loop.

>>> Loop(sv1, delay).each(param6, Loop(sv2, delay).each(sv3, delay)).run()

Supported commands to .each are:

    - Parameter: anything with a .get method and .name or .names see
      parameter.py for options
    - ActiveLoop
    - Task: any callable that does not generate data
    - Wait: a delay
"""
import logging
import time
from datetime import datetime
from typing import Optional, Sequence

import numpy as np

from qcodes.data.data_array import DataArray
from qcodes.data.data_set import new_data
from qcodes.station import Station
from qcodes.utils.helpers import full_class, tprint, wait_secs
from qcodes.utils.metadata import Metadatable

from .actions import (
    BreakIf,
    Task,
    Wait,
    _actions_snapshot,
    _Measure,
    _Nest,
    _QcodesBreak,
)

log = logging.getLogger(__name__)


def active_loop():
    return ActiveLoop.active_loop


def active_data_set():
    loop = active_loop()
    if loop is not None and loop.data_set is not None:
        return loop.data_set
    else:
        return None


class Loop(Metadatable):
    """
    The entry point for creating measurement loops

    Args:
        sweep_values: a SweepValues or compatible object describing what
            parameter to set in the loop and over what values
        delay: a number of seconds to wait after setting a value before
            continuing. 0 (default) means no waiting and no warnings. > 0
            means to wait, potentially filling the delay time with monitoring,
            and give an error if you wait longer than expected.
        progress_interval: should progress of the loop every x seconds. Default
            is None (no output)

    After creating a Loop, you attach one or more ``actions`` to it, making an
    ``ActiveLoop``

    ``actions`` is a sequence of things to do at each ``Loop`` step: that can be
    a ``Parameter`` to measure, a ``Task`` to do (any callable that does not
    yield data), ``Wait`` times, or another ``ActiveLoop`` or ``Loop`` to nest
    inside this one.
    """
    def __init__(self, sweep_values, delay=0, station=None,
                 progress_interval=None):
        super().__init__()
        if delay < 0:
            raise ValueError(f"delay must be > 0, not {repr(delay)}")

        self.sweep_values = sweep_values
        self.delay = delay
        self.station = station
        self.nested_loop = None
        self.actions = None
        self.then_actions = ()
        self.bg_task = None
        self.bg_final_task = None
        self.bg_min_delay = None
        self.progress_interval = progress_interval

    def __getitem__(self, item):
        """
        Retrieves action with index `item`
        Args:
            item: actions index

        Returns:
            loop.actions[item]
        """
        return self.actions[item]

    def loop(self, sweep_values, delay=0):
        """
        Nest another loop inside this one.

        Args:
            sweep_values:
            delay (int):

        Examples:
            >>> Loop(sv1, d1).loop(sv2, d2).each(*a)

            is equivalent to:

            >>> Loop(sv1, d1).each(Loop(sv2, d2).each(*a))

        Returns: a new Loop object - the original is untouched
        """
        out = self._copy()

        if out.nested_loop:
            # nest this new loop inside the deepest level
            out.nested_loop = out.nested_loop.loop(sweep_values, delay)
        else:
            out.nested_loop = Loop(sweep_values, delay)

        return out

    def _copy(self):
        out = Loop(self.sweep_values, self.delay,
                   progress_interval=self.progress_interval)
        out.nested_loop = self.nested_loop
        out.then_actions = self.then_actions
        out.station = self.station
        return out

    def each(self, *actions):
        """
        Perform a set of actions at each setting of this loop.
        TODO(setting vs setpoints) ? better be verbose.

        Args:
            *actions (Any): actions to perform at each setting of the loop

        Each action can be:

        - a Parameter to measure
        - a Task to execute
        - a Wait
        - another Loop or ActiveLoop

        """
        actions = list(actions)

        self.validate_actions(*actions)

        if self.nested_loop:
            # recurse into the innermost loop and apply these actions there
            actions = [self.nested_loop.each(*actions)]

        return ActiveLoop(self.sweep_values, self.delay, *actions,
                          then_actions=self.then_actions, station=self.station,
                          progress_interval=self.progress_interval,
                          bg_task=self.bg_task, bg_final_task=self.bg_final_task, bg_min_delay=self.bg_min_delay)

    def with_bg_task(self, task, bg_final_task=None, min_delay=0.01):
        """
        Attaches a background task to this loop.

        Args:
            task: A callable object with no parameters. This object will be
                invoked periodically during the measurement loop.

            bg_final_task: A callable object with no parameters. This object will be
                invoked to clean up after or otherwise finish the background
                task work.

            min_delay (int, float): The minimum number of seconds to wait
                between task invocations. Defaults to 0.01 s.
                Note that if a task is doing a lot of processing it is recommended
                to increase min_delay.
                Note that the actual time between task invocations may be much
                longer than this, as the task is only run between passes
                through the loop.
        """
        return _attach_bg_task(self, task, bg_final_task, min_delay)

    @staticmethod
    def validate_actions(*actions):
        """
        Whitelist acceptable actions, so we can give nice error messages
        if an action is not recognized
        """
        for action in actions:
            if isinstance(action, (Task, Wait, BreakIf, ActiveLoop)):
                continue
            if hasattr(action, 'get') and (hasattr(action, 'name') or
                                           hasattr(action, 'names')):
                continue
            raise TypeError('Unrecognized action:', action,
                            'Allowed actions are: objects (parameters) with '
                            'a `get` method and `name` or `names` attribute, '
                            'and `Task`, `Wait`, `BreakIf`, and `ActiveLoop` '
                            'objects. `Loop` objects are OK too, except in '
                            'Station default measurements.')


    def then(self, *actions, overwrite=False):
        """
        Attach actions to be performed after the loop completes.

        These can only be ``Task`` and ``Wait`` actions, as they may not generate
        any data.

        returns a new Loop object - the original is untouched

        This is more naturally done to an ActiveLoop (ie after .each())
        and can also be done there, but it's allowed at this stage too so that
        you can define final actions and share them among several ``Loops`` that
        have different loop actions, or attach final actions to a Loop run

        TODO:
            examples of this ? with default actions.

        Args:
            *actions: ``Task`` and ``Wait`` objects to execute in order

            overwrite: (default False) whether subsequent .then() calls (including
                calls in an ActiveLoop after .then() has already been called on
                the Loop) will add to each other or overwrite the earlier ones.
        Returns:
            a new Loop object - the original is untouched
        """
        return _attach_then_actions(self._copy(), actions, overwrite)

    def snapshot_base(self, update: Optional[bool] = False,
                      params_to_skip_update: Optional[Sequence[str]] = None):
        """
        State of the loop as a JSON-compatible dict (everything that
        the custom JSON encoder class :class:'qcodes.utils.helpers.NumpyJSONEncoder'
        supports).

        Args:
            update: If True, update the state by querying the underlying
                sweep_values and actions. If None only update state if known
                to be invalid. If False, just use the latest values
                in memory.
            params_to_skip_update: Unused in this implementation.

        Returns:
            dict: base snapshot
        """
        return {
            '__class__': full_class(self),
            'sweep_values': self.sweep_values.snapshot(update=update),
            'delay': self.delay,
            'then_actions': _actions_snapshot(self.then_actions, update)
        }


def _attach_then_actions(loop, actions, overwrite):
    """Inner code for both Loop.then and ActiveLoop.then."""
    for action in actions:
        if not isinstance(action, (Task, Wait)):
            raise TypeError('Unrecognized action:', action,
                            '.then() allows only `Task` and `Wait` '
                            'actions.')

    if overwrite:
        loop.then_actions = actions
    else:
        loop.then_actions = loop.then_actions + actions

    return loop


def _attach_bg_task(loop, task, bg_final_task, min_delay):
    """Inner code for both Loop and ActiveLoop.bg_task"""
    if loop.bg_task is None:
        loop.bg_task = task
        loop.bg_min_delay = min_delay
    else:
        raise RuntimeError('Only one background task is allowed per loop')

    if bg_final_task:
        loop.bg_final_task = bg_final_task

    return loop


class ActiveLoop(Metadatable):
    """
    Created by attaching ``actions`` to a ``Loop``, this is the object that
    actually runs a measurement loop. An ``ActiveLoop`` can no longer be nested,
    only run, or used as an action inside another ``Loop`` which will run the
    whole thing.

    The ``ActiveLoop`` determines what ``DataArrays`` it will need to hold the
    data it collects, and it creates a ``DataSet`` holding these ``DataArrays``
    """

    # Currently active loop, is set when calling loop.run(set_active=True)
    # is reset to None when active measurement is finished
    active_loop = None

    def __init__(self, sweep_values, delay, *actions, then_actions=(),
                 station=None, progress_interval=None, bg_task=None,
                 bg_final_task=None, bg_min_delay=None):
        super().__init__()
        self.sweep_values = sweep_values
        self.delay = delay
        self.actions = list(actions)
        self.progress_interval = progress_interval
        self.then_actions = then_actions
        self.station = station
        self.bg_task = bg_task
        self.bg_final_task = bg_final_task
        self.bg_min_delay = bg_min_delay
        self.data_set = None

        # if the first action is another loop, it changes how delays
        # happen - the outer delay happens *after* the inner var gets
        # set to its initial value
        self._nest_first = hasattr(actions[0], 'containers')

    def __getitem__(self, item):
        """
        Retrieves action with index `item`
        Args:
            item: actions index

        Returns:
            loop.actions[item]
        """
        return self.actions[item]

    def then(self, *actions, overwrite=False):
        """
        Attach actions to be performed after the loop completes.

        These can only be ``Task`` and ``Wait`` actions, as they may not
        generate any data.

        returns a new ActiveLoop object - the original is untouched



        Args:
            *actions: ``Task`` and ``Wait`` objects to execute in order

            overwrite: (default False) whether subsequent .then() calls (including
                calls in an ActiveLoop after .then() has already been called on
                the Loop) will add to each other or overwrite the earlier ones.
        """
        loop = ActiveLoop(self.sweep_values, self.delay, *self.actions,
                          then_actions=self.then_actions, station=self.station)
        return _attach_then_actions(loop, actions, overwrite)

    def with_bg_task(self, task, bg_final_task=None, min_delay=0.01):
        """
        Attaches a background task to this loop.

        Args:
            task: A callable object with no parameters. This object will be
                invoked periodically during the measurement loop.

            bg_final_task: A callable object with no parameters. This object will be
                invoked to clean up after or otherwise finish the background
                task work.

            min_delay (int, float): The minimum number of seconds to wait
                between task invocations. Note that the actual time between
                task invocations may be much longer than this, as the task is
                only run between passes through the loop. Defaults to 0.01 s.
        """
        return _attach_bg_task(self, task, bg_final_task, min_delay)

    def snapshot_base(self, update=False,
                      params_to_skip_update: Optional[Sequence[str]] = None):
        """Snapshot of this ActiveLoop's definition."""
        return {
            '__class__': full_class(self),
            'sweep_values': self.sweep_values.snapshot(update=update),
            'delay': self.delay,
            'actions': _actions_snapshot(self.actions, update),
            'then_actions': _actions_snapshot(self.then_actions, update)
        }

    def containers(self):
        """
        Finds the data arrays that will be created by the actions in this
        loop, and nests them inside this level of the loop.

        Recursively calls `.containers` on any enclosed actions.
        """
        loop_size = len(self.sweep_values)
        data_arrays = []
        loop_array = DataArray(parameter=self.sweep_values.parameter,
                               is_setpoint=True)
        loop_array.nest(size=loop_size)

        data_arrays = [loop_array]
        # hack set_data into actions
        new_actions = self.actions[:]
        if hasattr(self.sweep_values, "parameters"): # combined parameter
            for parameter in self.sweep_values.parameters:
                new_actions.append(parameter)

        for i, action in enumerate(new_actions):
            if hasattr(action, 'containers'):
                action_arrays = action.containers()

            elif hasattr(action, 'get'):
                # this action is a parameter to measure
                # note that this supports lists (separate output arrays)
                # and arrays (nested in one/each output array) of return values
                action_arrays = self._parameter_arrays(action)

            else:
                # this *is* covered but the report misses it because Python
                # optimizes it away. See:
                # https://bitbucket.org/ned/coveragepy/issues/198
                continue  # pragma: no cover

            for array in action_arrays:
                array.nest(size=loop_size, action_index=i,
                           set_array=loop_array)
            data_arrays.extend(action_arrays)

        return data_arrays

    def _parameter_arrays(self, action):
        out = []

        # first massage all the input parameters to the general multi-name form
        if hasattr(action, 'names'):
            names = action.names
            full_names = action.full_names
            labels = getattr(action, 'labels', names)
            if len(labels) != len(names):
                raise ValueError('must have equal number of names and labels')
            action_indices = tuple((i,) for i in range(len(names)))
        elif hasattr(action, 'name'):
            names = (action.name,)
            full_names = (action.full_name,)
            labels = (getattr(action, 'label', action.name),)
            action_indices = ((),)
        else:
            raise ValueError('a gettable parameter must have .name or .names')
        if hasattr(action, 'names') and hasattr(action, 'units'):
            units = action.units
        elif hasattr(action, 'unit'):
            units = (action.unit,)
        else:
            units = tuple(['']*len(names))
        num_arrays = len(names)
        shapes = getattr(action, 'shapes', None)
        sp_vals = getattr(action, 'setpoints', None)
        sp_names = getattr(action, 'setpoint_names', None)
        sp_labels = getattr(action, 'setpoint_labels', None)
        sp_units = getattr(action, 'setpoint_units', None)

        if shapes is None:
            shapes = (getattr(action, 'shape', ()),) * num_arrays
            sp_vals = (sp_vals,) * num_arrays
            sp_names = (sp_names,) * num_arrays
            sp_labels = (sp_labels,) * num_arrays
            sp_units = (sp_units,) * num_arrays
        else:
            sp_blank = (None,) * num_arrays
            # _fill_blank both supplies defaults and tests length
            # if values are supplied (for shapes it ONLY tests length)
            shapes = self._fill_blank(shapes, sp_blank)
            sp_vals = self._fill_blank(sp_vals, sp_blank)
            sp_names = self._fill_blank(sp_names, sp_blank)
            sp_labels = self._fill_blank(sp_labels, sp_blank)
            sp_units = self._fill_blank(sp_units, sp_blank)

        # now loop through these all, to make the DataArrays
        # record which setpoint arrays we've made, so we don't duplicate
        all_setpoints = {}
        for name, full_name, label, unit, shape, i, sp_vi, sp_ni, sp_li, sp_ui in zip(
                names, full_names, labels, units, shapes, action_indices,
                sp_vals, sp_names, sp_labels, sp_units):

            if shape is None or shape == ():
                shape, sp_vi, sp_ni, sp_li, sp_ui= (), (), (), (), ()
            else:
                sp_blank = (None,) * len(shape)
                sp_vi = self._fill_blank(sp_vi, sp_blank)
                sp_ni = self._fill_blank(sp_ni, sp_blank)
                sp_li = self._fill_blank(sp_li, sp_blank)
                sp_ui = self._fill_blank(sp_ui, sp_blank)

            setpoints = ()
            # loop through dimensions of shape to make the setpoint arrays
            for j, (vij, nij, lij, uij) in enumerate(zip(sp_vi, sp_ni, sp_li, sp_ui)):
                sp_def = (shape[: 1 + j], j, setpoints, vij, nij, lij, uij)
                if sp_def not in all_setpoints:
                    all_setpoints[sp_def] = self._make_setpoint_array(*sp_def)
                    out.append(all_setpoints[sp_def])
                setpoints = setpoints + (all_setpoints[sp_def],)

            # finally, make the output data array with these setpoints
            out.append(DataArray(name=name, full_name=full_name, label=label,
                                 shape=shape, action_indices=i, unit=unit,
                                 set_arrays=setpoints, parameter=action))

        return out

    def _fill_blank(self, inputs, blanks):
        if inputs is None:
            return blanks
        elif len(inputs) == len(blanks):
            return inputs
        else:
            raise ValueError('Wrong number of inputs supplied')

    def _make_setpoint_array(self, shape, i, prev_setpoints, vals, name,
                             label, unit):
        if vals is None:
            vals = self._default_setpoints(shape)
        elif isinstance(vals, DataArray):
            # can't simply use the DataArray, even though that's
            # what we're going to return here, because it will
            # get nested (don't want to alter the original)
            # DataArrays do have the advantage though of already including
            # name and label, so take these if they exist
            if vals.name is not None:
                name = vals.name
            if vals.label is not None:
                label = vals.label

            # extract a copy of the numpy array
            vals = np.array(vals.ndarray)
        else:
            # turn any sequence into a (new) numpy array
            vals = np.array(vals)

        if vals.shape != shape:
            raise ValueError('nth setpoint array should have shape matching '
                             'the first n dimensions of shape.')

        if name is None:
            name = f'index{i}'

        return DataArray(name=name, label=label, set_arrays=prev_setpoints,
                         shape=shape, preset_data=vals, unit=unit, is_setpoint=True)

    def _default_setpoints(self, shape):
        if len(shape) == 1:
            return np.arange(0, shape[0], 1)

        sp = np.ndarray(shape)
        sp_inner = self._default_setpoints(shape[1:])
        for i in range(len(sp)):
            sp[i] = sp_inner

        return sp

    def set_common_attrs(self, data_set, use_threads):
        """
        set a couple of common attributes that the main and nested loops
        all need to have:
        - the DataSet collecting all our measurements
        - a queue for communicating with the main process
        """
        self.data_set = data_set
        self.use_threads = use_threads
        for action in self.actions:
            if hasattr(action, 'set_common_attrs'):
                action.set_common_attrs(data_set, use_threads)

    def get_data_set(self, *args, **kwargs):
        """
        Return the data set for this loop.

        If no data set has been created yet, a new one will be created and
        returned. Note that all arguments can only be provided when the
        `DataSet` is first created; giving these during `run` when
        `get_data_set` has already been called on its own is an error.

        Args:
            data_manager: a DataManager instance (omit to use default,
                False to store locally)

        kwargs are passed along to data_set.new_data. The key ones are:

        Args:
            location: the location of the DataSet, a string whose meaning
                depends on formatter and io, or False to only keep in memory.
                May be a callable to provide automatic locations. If omitted, will
                use the default DataSet.location_provider
            name: if location is default or another provider function, name is
                a string to add to location to make it more readable/meaningful
                to users
            formatter: knows how to read and write the file format
                default can be set in DataSet.default_formatter
            io: knows how to connect to the storage (disk vs cloud etc)
            write_period: how often to save to storage during the loop.
                default 5 sec, use None to write only at the end

        returns:
            a DataSet object that we can use to plot
        """
        if self.data_set is None:
            data_set = new_data(arrays=self.containers(), *args, **kwargs)
            self.data_set = data_set

        else:
            has_args = len(kwargs) or len(args)
            if has_args:
                raise RuntimeError(
                    'The DataSet for this loop already exists. '
                    'You can only provide DataSet attributes, such as '
                    'data_manager, location, name, formatter, io, '
                    'write_period, when the DataSet is first created.')

        return self.data_set

    def run_temp(self, **kwargs):
        """
        wrapper to run this loop in the foreground as a temporary data set,
        especially for use in composite parameters that need to run a Loop
        as part of their get method
        """
        return self.run(quiet=True, location=False, **kwargs)

    def run(self, use_threads=False, quiet=False, station=None,
            progress_interval=False, set_active=True, *args, **kwargs):
        """
        Execute this loop.

        Args:
            use_threads: (default False): whenever there are multiple `get` calls
                back-to-back, execute them in separate threads so they run in
                parallel (as long as they don't block each other)
            quiet: (default False): set True to not print anything except errors
            station: a Station instance for snapshots (omit to use a previously
                provided Station, or the default Station)
            progress_interval (int, float): show progress of the loop every x
                seconds. If provided here, will override any interval provided
                with the Loop definition. Defaults to None

        kwargs are passed along to data_set.new_data. These can only be
        provided when the `DataSet` is first created; giving these during `run`
        when `get_data_set` has already been called on its own is an error.
        The key ones are:

        Args:
            location: the location of the DataSet, a string whose meaning
                depends on formatter and io, or False to only keep in memory.
                May be a callable to provide automatic locations. If omitted, will
                use the default DataSet.location_provider
            name: if location is default or another provider function, name is
                a string to add to location to make it more readable/meaningful
                to users
            formatter: knows how to read and write the file format
                default can be set in DataSet.default_formatter
            io: knows how to connect to the storage (disk vs cloud etc)
                write_period: how often to save to storage during the loop.
                default 5 sec, use None to write only at the end


        returns:
            a DataSet object that we can use to plot
        """
        if progress_interval is not False:
            self.progress_interval = progress_interval

        data_set = self.get_data_set(*args, **kwargs)

        self.set_common_attrs(data_set=data_set, use_threads=use_threads)

        station = station or self.station or Station.default
        if station:
            data_set.add_metadata({'station': station.snapshot()})

        # information about the loop definition is in its snapshot
        data_set.add_metadata({'loop': self.snapshot()})
        # then add information about how and when it was run
        ts = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        data_set.add_metadata({'loop': {
            'ts_start': ts,
            'use_threads': use_threads,
        }})

        data_set.save_metadata()

        if set_active:
            ActiveLoop.active_loop = self

        try:
            if not quiet:
                print(datetime.now().strftime('Started at %Y-%m-%d %H:%M:%S'))
            self._run_wrapper()
            ds = self.data_set
        finally:
            if not quiet:
                print(repr(self.data_set))
                print(datetime.now().strftime('Finished at %Y-%m-%d %H:%M:%S'))

            # After normal loop execution we clear the data_set so we can run
            # again. But also if something went wrong during the loop execution
            # we want to clear the data_set attribute so we don't try to reuse
            # this one later.
            self.data_set = None
            if set_active:
                ActiveLoop.active_loop = None

        return ds

    def _compile_actions(self, actions, action_indices=()):
        callables = []
        measurement_group = []
        for i, action in enumerate(actions):
            new_action_indices = action_indices + (i,)
            if hasattr(action, 'get'):
                measurement_group.append((action, new_action_indices))
                continue
            elif measurement_group:
                callables.append(_Measure(measurement_group, self.data_set,
                                          self.use_threads))
                measurement_group[:] = []

            callables.append(self._compile_one(action, new_action_indices))

        if measurement_group:
            callables.append(_Measure(measurement_group, self.data_set,
                                      self.use_threads))
            measurement_group[:] = []

        return callables

    def _compile_one(self, action, new_action_indices):
        if isinstance(action, Wait):
            return Task(self._wait, action.delay)
        elif isinstance(action, ActiveLoop):
            return _Nest(action, new_action_indices)
        else:
            return action

    def _run_wrapper(self, *args, **kwargs):
        try:
            self._run_loop(*args, **kwargs)
        finally:
            if hasattr(self, 'data_set'):
                # TODO (giulioungaretti) WTF?
                # somehow this does not show up in the data_set returned by
                # run(), but it is saved to the metadata
                ts = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
                self.data_set.add_metadata({'loop': {'ts_end': ts}})
                self.data_set.finalize()

    def _run_loop(self, first_delay=0, action_indices=(),
                  loop_indices=(), current_values=(),
                  **ignore_kwargs):
        """
        the routine that actually executes the loop, and can be called
        from one loop to execute a nested loop

        first_delay: any delay carried over from an outer loop
        action_indices: where we are in any outer loop action arrays
        loop_indices: setpoint indices in any outer loops
        current_values: setpoint values in any outer loops
        signal_queue: queue to communicate with main process directly
        ignore_kwargs: for compatibility with other loop tasks
        """

        # at the beginning of the loop, the time to wait after setting
        # the loop parameter may be increased if an outer loop requested longer
        delay = max(self.delay, first_delay)

        callables = self._compile_actions(self.actions, action_indices)
        n_callables = 0
        for item in callables:
            if hasattr(item, 'param_ids'):
                n_callables += len(item.param_ids)
            else:
                n_callables += 1
        t0 = time.time()
        last_task = t0
        imax = len(self.sweep_values)

        self.last_task_failed = False

        for i, value in enumerate(self.sweep_values):
            if self.progress_interval is not None:
                tprint('loop %s: %d/%d (%.1f [s])' % (
                    self.sweep_values.name, i, imax, time.time() - t0),
                    dt=self.progress_interval, tag='outerloop')
                if i:
                    tprint("Estimated finish time: %s" % (
                        time.asctime(time.localtime(t0 + ((time.time() - t0) * imax / i)))),
                           dt=self.progress_interval, tag="finish")

            set_val = self.sweep_values.set(value)

            new_indices = loop_indices + (i,)
            new_values = current_values + (value,)
            data_to_store = {}

            if hasattr(self.sweep_values, "parameters"):  # combined parameter
                set_name = self.data_set.action_id_map[action_indices]
                if hasattr(self.sweep_values, 'aggregate'):
                    value = self.sweep_values.aggregate(*set_val)
                # below is useful but too verbose even at debug
                # log.debug('Calling .store method of DataSet because '
                #           'sweep_values.parameters exist')
                self.data_set.store(new_indices, {set_name: value})
                # set_val list of values to set [param1_setpoint, param2_setpoint ..]
                for j, val in enumerate(set_val):
                    set_index = action_indices + (j+n_callables, )
                    set_name = (self.data_set.action_id_map[set_index])
                    data_to_store[set_name] = val
            else:
                set_name = self.data_set.action_id_map[action_indices]
                data_to_store[set_name] = value
            # below is useful but too verbose even at debug
            # log.debug('Calling .store method of DataSet because a sweep step'
            #           ' was taken')
            self.data_set.store(new_indices, data_to_store)

            if not self._nest_first:
                # only wait the delay time if an inner loop will not inherit it
                self._wait(delay)

            try:
                for f in callables:
                    # below is useful but too verbose even at debug
                    # log.debug('Going through callables at this sweep step.'
                    #           ' Calling {}'.format(f))
                    f(first_delay=delay,
                      loop_indices=new_indices,
                      current_values=new_values)

                    # after the first action, no delay is inherited
                    delay = 0
            except _QcodesBreak:
                break

            # after the first setpoint, delay reverts to the loop delay
            delay = self.delay

            # now check for a background task and execute it if it's
            # been long enough since the last time
            # don't let exceptions in the background task interrupt
            # the loop
            # if the background task fails twice consecutively, stop
            # executing it
            if self.bg_task is not None:
                t = time.time()
                if t - last_task >= self.bg_min_delay:
                    try:
                        self.bg_task()
                    except Exception:
                        if self.last_task_failed:
                            self.bg_task = None
                        self.last_task_failed = True
                        log.exception("Failed to execute bg task")

                    last_task = t

        # run the background task one last time to catch the last setpoint(s)
        if self.bg_task is not None:
            log.debug('Running the background task one last time.')
            self.bg_task()

        # the loop is finished - run the .then actions
        #log.debug('Finishing loop, running the .then actions...')
        for f in self._compile_actions(self.then_actions, ()):
            #log.debug('...running .then action {}'.format(f))
            f()

        # run the bg_final_task from the bg_task:
        if self.bg_final_task is not None:
            log.debug('Running the bg_final_task')
            self.bg_final_task()

    def _wait(self, delay):
        if delay:
            finish_clock = time.perf_counter() + delay
            t = wait_secs(finish_clock)
            time.sleep(t)




# This file was generated by 'versioneer.py' (0.21) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2021-10-25T11:01:43+0200",
 "dirty": false,
 "error": null,
 "full-revisionid": "c3b3749c1fd9e5171da9313a9a630a3b54812700",
 "version": "0.30.0.dev0"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



"""
Tests for `qcodes.utils.logger`.
"""
import logging
import os
from copy import copy

import pytest
from packaging import version

import qcodes as qc
import qcodes.logger as logger
from qcodes.logger.log_analysis import capture_dataframe

TEST_LOG_MESSAGE = 'test log message'

pytest_version = version.parse(pytest.__version__)
assert isinstance(pytest_version, version.Version)

if pytest_version.major >= 6:
    NUM_PYTEST_LOGGERS = 2
else:
    NUM_PYTEST_LOGGERS = 1


@pytest.fixture
def remove_root_handlers():
    root_logger = logging.getLogger()
    handlers = copy(root_logger.handlers)
    for handler in handlers:
        handler.close()
        root_logger.removeHandler(handler)
    logger.logger.file_handler = None
    logger.logger.console_handler = None


@pytest.fixture
def awg5208():

    import qcodes.instrument.sims as sims
    from qcodes.instrument_drivers.tektronix.AWG5208 import AWG5208
    visalib = sims.__file__.replace('__init__.py',
                                    'Tektronix_AWG5208.yaml@sim')

    logger.start_logger()

    inst = AWG5208('awg_sim',
                   address='GPIB0::1::INSTR',
                   visalib=visalib)

    try:
        yield inst
    finally:
        inst.close()


@pytest.fixture
def model372():
    import qcodes.instrument.sims as sims
    from qcodes.tests.drivers.test_lakeshore import Model_372_Mock

    logger.LOGGING_SEPARATOR = ' - '

    logger.start_logger()

    visalib = sims.__file__.replace('__init__.py',
                                    'lakeshore_model372.yaml@sim')

    inst = Model_372_Mock('lakeshore_372', 'GPIB::3::INSTR',
                          visalib=visalib, device_clear=False)
    inst.sample_heater.range_limits([0, 0.25, 0.5, 1, 2, 3, 4, 7])
    inst.warmup_heater.range_limits([0, 0.25, 0.5, 1, 2, 3, 4, 7])
    try:
        yield inst
    finally:
        inst.close()


@pytest.fixture()
def AMI430_3D():
    import numpy as np

    import qcodes.instrument.sims as sims
    from qcodes.instrument.ip_to_visa import AMI430_VISA
    from qcodes.instrument_drivers.american_magnetics.AMI430 import AMI430_3D
    visalib = sims.__file__.replace('__init__.py', 'AMI430.yaml@sim')
    mag_x = AMI430_VISA('x', address='GPIB::1::INSTR', visalib=visalib,
                        terminator='\n', port=1)
    mag_y = AMI430_VISA('y', address='GPIB::2::INSTR', visalib=visalib,
                        terminator='\n', port=1)
    mag_z = AMI430_VISA('z', address='GPIB::3::INSTR', visalib=visalib,
                        terminator='\n', port=1)
    field_limit = [
        lambda x, y, z: x == 0 and y == 0 and z < 3,
        lambda x, y, z: np.linalg.norm([x, y, z]) < 2
    ]
    driver = AMI430_3D("AMI430-3D", mag_x, mag_y, mag_z, field_limit)
    try:
        yield driver, mag_x, mag_y, mag_z
    finally:
        driver.close()
        mag_x.close()
        mag_y.close()
        mag_z.close()


def test_get_log_file_name():
    fp = logger.logger.get_log_file_name().split(os.sep)
    assert str(os.getpid()) in fp[-1]
    assert logger.logger.PYTHON_LOG_NAME in fp[-1]
    assert fp[-2] == logger.logger.LOGGING_DIR
    assert fp[-3] == '.qcodes'


@pytest.mark.usefixtures("remove_root_handlers")
def test_start_logger():
    # remove all Handlers
    logger.start_logger()
    assert isinstance(logger.get_console_handler(), logging.Handler)
    assert isinstance(logger.get_file_handler(), logging.Handler)

    console_level = logger.get_level_code(qc.config.logger.console_level)
    file_level = logger.get_level_code(qc.config.logger.file_level)
    assert logger.get_console_handler().level == console_level
    assert logger.get_file_handler().level == file_level

    assert logging.getLogger().level == logger.get_level_code('DEBUG')


@pytest.mark.usefixtures("remove_root_handlers")
def test_start_logger_twice():
    logger.start_logger()
    logger.start_logger()
    handlers = logging.getLogger().handlers
    # there is one or two loggers registered from pytest
    # depending on the version
    # and the telemetry logger is always off in the tests
    assert len(handlers) == 2+NUM_PYTEST_LOGGERS


@pytest.mark.usefixtures("remove_root_handlers")
def test_set_level_without_starting_raises():
    with pytest.raises(RuntimeError):
        with logger.console_level('DEBUG'):
            pass
    # there is one or two loggers registered from pytest
    # depending on the version
    assert len(logging.getLogger().handlers) == NUM_PYTEST_LOGGERS


@pytest.mark.usefixtures("remove_root_handlers")
def test_handler_level():
    with logger.LogCapture(level=logging.INFO) as logs:
        logging.debug(TEST_LOG_MESSAGE)
    assert logs.value == ''

    with logger.LogCapture(level=logging.INFO) as logs:
        with logger.handler_level(level=logging.DEBUG,
                                  handler=logs.string_handler):
            print(logs.string_handler)
            logging.debug(TEST_LOG_MESSAGE)
    assert logs.value.strip() == TEST_LOG_MESSAGE


@pytest.mark.usefixtures("remove_root_handlers")
def test_filter_instrument(AMI430_3D):

    driver, mag_x, mag_y, mag_z = AMI430_3D

    logger.start_logger()

    # filter one instrument
    driver.cartesian((0, 0, 0))
    with logger.LogCapture(level=logging.DEBUG) as logs:
        with logger.filter_instrument(mag_x, handler=logs.string_handler):
            driver.cartesian((0, 0, 1))
    for line in logs.value.splitlines():
        assert '[x(AMI430_VISA)]' in line
        assert '[y(AMI430_VISA)]' not in line
        assert '[z(AMI430_VISA)]' not in line

    # filter multiple instruments
    driver.cartesian((0, 0, 0))
    with logger.LogCapture(level=logging.DEBUG) as logs:
        with logger.filter_instrument((mag_x, mag_y), handler=logs.string_handler):
            driver.cartesian((0, 0, 1))

    any_x = False
    any_y = False
    for line in logs.value.splitlines():
        has_x = '[x(AMI430_VISA)]' in line
        has_y = '[y(AMI430_VISA)]' in line
        has_z = '[z(AMI430_VISA)]' in line

        assert has_x or has_y
        assert not has_z

        any_x |= has_x
        any_y |= has_y
    assert any_x
    assert any_y


@pytest.mark.usefixtures("remove_root_handlers")
def test_filter_without_started_logger_raises(AMI430_3D):

    driver, mag_x, mag_y, mag_z = AMI430_3D

    # filter one instrument
    driver.cartesian((0, 0, 0))
    with pytest.raises(RuntimeError):
        with logger.filter_instrument(mag_x):
            pass


@pytest.mark.usefixtures("remove_root_handlers")
def test_capture_dataframe():
    root_logger = logging.getLogger()
    with capture_dataframe() as (_, cb):
        root_logger.debug(TEST_LOG_MESSAGE)
        df = cb()
    assert len(df) == 1
    assert df.message[0] == TEST_LOG_MESSAGE


@pytest.mark.usefixtures("remove_root_handlers")
def test_channels(model372):
    """
    Test that messages logged in a channel are propagated to the
    main instrument.
    """
    inst = model372

    # set range to some other value so that it will actually be set in
    # the next call.
    inst.sample_heater.range_limits([0, 0.25, 0.5, 1, 2, 3, 4, 7])
    inst.sample_heater.set_range_from_temperature(1)
    with logger.LogCapture(level=logging.DEBUG) as logs_unfiltered:
        inst.sample_heater.set_range_from_temperature(0.1)

    # reset without capturing
    inst.sample_heater.set_range_from_temperature(1)
    # rerun with instrument filter
    with logger.LogCapture(level=logging.DEBUG) as logs_filtered,\
            logger.filter_instrument(inst,
                                     handler=logs_filtered.string_handler):
        inst.sample_heater.set_range_from_temperature(0.1)

    logs_filtered = [l for l in logs_filtered.value.splitlines()
                        if '[lakeshore' in l]
    logs_unfiltered = [l for l in logs_unfiltered.value.splitlines()
                        if '[lakeshore' in l]

    for f, u in zip(logs_filtered, logs_unfiltered):
        assert f == u


@pytest.mark.usefixtures("remove_root_handlers")
def test_channels_nomessages(model372):
    """
    Test that messages logged in a channel are not propagated to
    any instrument.
    """
    inst = model372
    # test with wrong instrument
    mock = qc.Instrument('mock')
    inst.sample_heater.set_range_from_temperature(1)
    with logger.LogCapture(level=logging.DEBUG) as logs,\
            logger.filter_instrument(mock, handler=logs.string_handler):
        inst.sample_heater.set_range_from_temperature(0.1)
    logs = [l for l in logs.value.splitlines()
            if '[lakeshore' in l]
    assert len(logs) == 0
    mock.close()


@pytest.mark.usefixtures("remove_root_handlers", "awg5208")
def test_instrument_connect_message():
    """
    Test that the connect_message method logs as expected

    This test kind of belongs both here and in the tests for the instrument
    code, but it is more conveniently written here
    """

    with open(logger.get_log_file_name(), encoding="utf-8") as f:
        lines = f.readlines()

    con_mssg_log_line = lines[-1]

    sep = logger.logger.LOGGING_SEPARATOR

    con_mss = con_mssg_log_line.split(sep)[-1]
    idn = {"vendor": "QCoDeS", "model": "AWG5208",
           "serial": "1000", "firmware": "0.1"}
    expected_con_mssg = ("[awg_sim(AWG5208)] Connected to instrument: "
                         f"{idn}\n")

    assert con_mss == expected_con_mssg


@pytest.mark.usefixtures("remove_root_handlers")
def test_installation_info_logging():
    """
    Test that installation information is logged upon starting the logging
    """
    logger.start_logger()

    with open(logger.get_log_file_name()) as f:
        lines = f.readlines()

    assert "QCoDeS version:" in lines[-4]
    assert "QCoDeS installed in editable mode:" in lines[-3]
    assert "QCoDeS requirements versions:" in lines[-2]
    assert "All installed package versions:" in lines[-1]



import os

import pytest

from qcodes.instrument_drivers.tektronix.AWG5208 import AWG5208
import qcodes.instrument.sims as sims
visalib = sims.__file__.replace('__init__.py', 'Tektronix_AWG5208.yaml@sim')


@pytest.fixture(scope='function')
def awg():
    awg_sim = AWG5208('awg_sim',
                      address='GPIB0::1::INSTR',
                      visalib=visalib)
    yield awg_sim

    awg_sim.close()


def test_init_awg(awg):

    idn_dict = awg.IDN()

    assert idn_dict['vendor'] == 'QCoDeS'


def test_channel_resolution_docstring(awg):

    expected_docstring = ("12 bit resolution allows for four "
                          "markers, 13 bit resolution "
                          "allows for three, etc. with 16 bit "
                          "allowing for ZERO markers")

    actual_docstring = awg.ch1.resolution.__doc__.split(os.linesep)[0]

    assert actual_docstring == expected_docstring



import logging
from collections import namedtuple
from operator import attrgetter
from traceback import format_exc
from typing import TYPE_CHECKING, Set

if TYPE_CHECKING:
    from .data_set import DataSet


log = logging.getLogger(__name__)
ArrayGroup = namedtuple("ArrayGroup", "shape set_arrays data name")


class Formatter:
    """
    Data file formatters

    Formatters translate between DataSets and data files.

    Each Formatter is expected to implement writing methods:

    - ``write``: to write the ``DataArrays``
    - ``write_metadata``: to write the metadata structure

    Optionally, if this Formatter keeps the data file(s) open
    between write calls, it may implement:

    - ``close_file``: to perform any final cleanup and release the
      file and any other resources.

    and reading methods:

    - ``read`` or ``read_one_file`` to reconstruct the ``DataArrays``, either
      all at once (``read``) or one file at a time, supplied by the base class
      ``read`` method that loops over all data files at the correct location.

    - ``read_metadata``: to reload saved metadata. If a subclass overrides
      ``read``, this method should call ``read_metadata``, but keep it also
      as a separate method because it occasionally gets called independently.

    All of these methods accept a ``data_set`` argument, which should be a
    ``DataSet`` object. Even if you are loading a new data set from disk, this
    object should already have attributes:

        - io: an IO manager (see qcodes.data.io)
          location: a string, like a file path, that identifies the DataSet and
          tells the IO manager where to store it
        - arrays: a dict of ``{array_id:DataArray}`` to read into.

    - read will create entries that don't yet exist.
    - write will write ALL DataArrays in the DataSet, using
      last_saved_index and modified_range, as well as whether or not
      it found the specified file, to determine how much to write.
    """

    ArrayGroup = ArrayGroup

    def write(self, data_set: 'DataSet', io_manager, location, write_metadata=True,
              force_write=False, only_complete=True):
        """
        Write the DataSet to storage.

        Subclasses must override this method.

        It is up to the Formatter to decide when to overwrite completely,
        and when to just append or otherwise update the file(s).

        Args:
            data_set: the data we are writing.
            io_manager (io_manager): base physical location to write to.
            location (str): the file location within the io_manager.
            write_metadata (bool): if True, then the metadata is written to disk
            force_write (bool): if True, then the data is written to disk
            only_complete (bool): Used only by the gnuplot formatter's
                overridden version of this method
        """
        raise NotImplementedError

    def read(self, data_set: 'DataSet') -> None:
        """
        Read the entire ``DataSet``.

        Find all files matching ``data_set.location`` (using io_manager.list)
        and call ``read_one_file`` on each. Subclasses may either override
        this method (if they use only one file or want to do their own
        searching) or override ``read_one_file`` to use the search and
        initialization functionality defined here.

        Args:
            data_set: the data to read into. Should already have
                attributes ``io`` (an io manager), ``location`` (string),
                and ``arrays`` (dict of ``{array_id: array}``, can be empty
                or can already have some or all of the arrays present, they
                expect to be overwritten)
        """
        io_manager = data_set.io
        location = data_set.location

        data_files = io_manager.list(location)
        if not data_files:
            raise OSError('no data found at ' + location)

        # in case the DataArrays exist but haven't been initialized
        for array in data_set.arrays.values():
            if array.ndarray is None:
                array.init_data()

        self.read_metadata(data_set)

        ids_read: Set[str] = set()
        for fn in data_files:
            with io_manager.open(fn, 'r') as f:
                try:
                    self.read_one_file(data_set, f, ids_read)
                except ValueError:
                    log.warning('error reading file ' + fn)
                    log.warning(format_exc())

    def write_metadata(self, data_set: 'DataSet',
                       io_manager, location, read_first=True, **kwargs):
        """
        Write the metadata for this DataSet to storage.

        Subclasses must override this method.

        Args:
            data_set: the data we are writing.
            io_manager (io_manager): base physical location to write to.
            location (str): the file location within the io_manager.
            read_first (Optional[bool]): whether to first look for previously
                saved metadata that may contain more information than the local
                copy.
        """
        raise NotImplementedError

    def read_metadata(self, data_set: 'DataSet'):
        """
        Read the metadata from this DataSet from storage.

        Subclasses must override this method.

        Args:
            data_set: the data to read metadata into
        """
        raise NotImplementedError

    def read_one_file(self, data_set: 'DataSet', f, ids_read):
        """
        Read data from a single file into a ``DataSet``.

        Formatter subclasses that break a DataSet into multiple data files may
        choose to override either this method, which handles one file at a
        time, or ``read`` which finds matching files on its own.

        Args:
            data_set: the data we are reading into.

            f: a file-like object to read from, as provided by
                ``io_manager.open``.

            ids_read (set): ``array_ids`` that we have already read.
                When you read an array, check that it's not in this set (except
                setpoints, which can be in several files with different inner
                loops) then add it to the set so other files know it should not
                be read again.

        Raises:
            ValueError: if a duplicate array_id of measured data is found
        """
        raise NotImplementedError

    def match_save_range(self, group, file_exists, only_complete=True):
        """
        Find the save range that will joins all changes in an array group.

        Matches all full-sized arrays: the data arrays plus the inner loop
        setpoint array.

        Note: if an outer loop has changed values (without the inner
        loop or measured data changing) we won't notice it here. We assume
        that before an iteration of the inner loop starts, the outer loop
        setpoint gets set and then does not change later.

        Args:
            group (Formatter.ArrayGroup): a ``namedtuple`` containing the
                arrays that go together in one file, as tuple ``group.data``.

            file_exists (bool): Does this file already exist? If True, and
                all arrays in the group agree on ``last_saved_index``, we
                assume the file has been written up to this index and we can
                append to it. Otherwise we will set the returned range to start
                from zero (so if the file does exist, it gets completely
                overwritten).

            only_complete (bool): Should we write all available new data,
                or only complete rows? If True, we write only the range of
                array indices which all arrays in the group list as modified,
                so that future writes will be able to do a clean append to
                the data file as more data arrives.
                Default True.

        Returns:
            Tuple(int, int): the first and last raveled indices that should
                be saved. Returns None if:
                    * no data is present
                    * no new data can be found
        """
        inner_setpoint = group.set_arrays[-1]
        full_dim_data = (inner_setpoint, ) + group.data

        # always return None if there are no modifications,
        # even if there are last_saved_index inconsistencies
        # so we don't do extra writing just to reshape the file
        for array in full_dim_data:
            if array.modified_range:
                break
        else:
            return None

        last_saved_index = inner_setpoint.last_saved_index

        if last_saved_index is None or not file_exists:
            if last_saved_index is None and file_exists:
                log.warning("Inconsistent file information. "
                            "last_save_index is None but file exists. "
                            "Will overwrite")
            if last_saved_index is not None and not file_exists:
                log.warning("Inconsistent file information. "
                            "last_save_index is not None but file does not "
                            "exist. Will rewrite from scratch")
            return self._match_save_range_whole_file(
                full_dim_data, only_complete)

        # force overwrite if inconsistent last_saved_index
        for array in group.data:
            if array.last_saved_index != last_saved_index:
                return self._match_save_range_whole_file(
                    full_dim_data, only_complete)

        return self._match_save_range_incremental(
            full_dim_data, last_saved_index, only_complete)

    @staticmethod
    def _match_save_range_whole_file(arrays, only_complete):
        max_save = None
        agg = (min if only_complete else max)
        for array in arrays:
            array_max = array.last_saved_index
            if array_max is None:
                array_max = -1
            mr = array.modified_range
            if mr:
                array_max = max(array_max, mr[1])
            max_save = (array_max if max_save is None else
                        agg(max_save, array_max))

        if max_save >= 0:
            return (0, max_save)
        else:
            return None

    @staticmethod
    def _match_save_range_incremental(arrays, last_saved_index, only_complete):
        mod_ranges = []
        for array in arrays:
            mr = array.modified_range
            if not mr:
                if only_complete:
                    return None
                else:
                    continue
            mod_ranges.append(mr)

        mod_range = mod_ranges[0]
        agg = (min if only_complete else max)
        for mr in mod_ranges[1:]:
            mod_range = (min(mod_range[0], mr[0]),
                         agg(mod_range[1], mr[1]))

        if last_saved_index >= mod_range[1]:
            return (0, last_saved_index)
        elif last_saved_index >= mod_range[0]:
            return (0, mod_range[1])
        else:
            return (last_saved_index + 1, mod_range[1])

    def group_arrays(self, arrays):
        """
        Find the sets of arrays which share all the same setpoint arrays.

        Some Formatters use this grouping to determine which arrays to save
        together in one file.

        Args:
            arrays (Dict[DataArray]): all the arrays in a DataSet

        Returns:
            List[Formatter.ArrayGroup]: namedtuples giving:

            - shape (Tuple[int]): dimensions as in numpy
            - set_arrays (Tuple[DataArray]): the setpoints of this group
            - data (Tuple[DataArray]): measured arrays in this group
            - name (str): a unique name of this group, obtained by joining
              the setpoint array ids.
        """

        set_array_sets = tuple({array.set_arrays
                                   for array in arrays.values()})
        all_set_arrays = set()
        for set_array_set in set_array_sets:
            all_set_arrays.update(set_array_set)

        grouped_data = [[] for _ in set_array_sets]

        for array in arrays.values():
            i = set_array_sets.index(array.set_arrays)
            if array not in all_set_arrays:  # array.set_arrays[-1] != array:
                # don't include the setpoint array itself in the data
                grouped_data[i].append(array)

        out = []
        id_getter = attrgetter('array_id')
        for set_arrays, data in zip(set_array_sets, grouped_data):
            leni = len(set_arrays)
            if not data and any(1 for other_set_arrays in set_array_sets if
                                len(other_set_arrays) > leni and
                                other_set_arrays[:leni] == set_arrays):
                # this is an outer loop that doesn't have any data of its own,
                # so skip it.
                # Inner-loop setpoints with no data is weird (we set values
                # but didn't measure anything there?) but we should keep it.
                continue

            group_name = '_'.join(sai.array_id for sai in set_arrays)
            out.append(self.ArrayGroup(shape=set_arrays[-1].shape,
                                       set_arrays=set_arrays,
                                       data=tuple(sorted(data, key=id_getter)),
                                       name=group_name))
        return out



import pytest
import timeit

import qcodes.instrument.sims as sims
from qcodes.instrument_drivers.tektronix.DPO7200xx import TektronixDPO7000xx

visalib = sims.__file__.replace('__init__.py', 'Tektronix_DPO7200xx.yaml@sim')


@pytest.fixture(scope='module')
def tektronix_dpo():
    """
    A six channel-per-relay instrument
    """
    driver = TektronixDPO7000xx(
        'dpo', address='TCPIP0::0.0.0.0::inst0::INSTR', visalib=visalib)

    yield driver
    driver.close()


def test_adjust_timer(tektronix_dpo):
    """
    After adjusting the type of the measurement or the source of the
    measurement, we need wait at least 0.1 seconds
    ('minimum_adjustment_time') before a measurement value can be
    retrieved. Test this.
    """

    timer = timeit.Timer(
        'tektronix_dpo.measurement[0].source1("CH1"),'
        'tektronix_dpo.measurement[0].amplitude()',
        globals=locals(),
    )
    min_time = tektronix_dpo.measurement[0]._minimum_adjustment_time
    repeats = timer.repeat(repeat=10, number=1)

    # The minimum time should be at least 95% of the 'minimum_adjustment_time'
    assert all(t > min_time * 0.95 for t in repeats)
    # To see why this fudge factor is necessary, try the following:
    # >>> import time
    # >>> import timeit
    # >>> timer = timeit.Timer("time.sleep(1E-3)")
    # >>> print(any(t < 1E-3 for t in timer.repeat(repeat=100, number=1)))
    # ... True
    # Conclusion: the command 'time.sleep(1E-3)' sometimes takes less
    # than 1E-3 seconds to return. Since the sleep time is not critical
    # to the microsecond, we don't care that we sometimes retrieve
    # measurements slightly sooner then 'minimum_adjustment_time'


def test_measurements_return_float(tektronix_dpo):
    amplitude = tektronix_dpo.measurement[0].amplitude()
    assert isinstance(amplitude, float)

    mean_amplitude = tektronix_dpo.measurement[0].amplitude.mean()
    assert isinstance(mean_amplitude, float)


def test_measurement_sets_state(tektronix_dpo):
    tektronix_dpo.measurement[1].frequency()
    assert tektronix_dpo.measurement[1].state() == 1



# ***************************************************************************
# *---------------------------- Attribute Defines ---------------------------*
# ***************************************************************************

IVI_ATTR_BASE = 1000000


IVI_INHERENT_ATTR_BASE = (
    IVI_ATTR_BASE + 50000
)  # base for inherent capability attributes

# base for IVI-defined class attributes
IVI_CLASS_ATTR_BASE = IVI_ATTR_BASE + 250000

# base for IviLxiSync attributes
IVI_LXISYNC_ATTR_BASE = IVI_ATTR_BASE + 950000


IVI_SPECIFIC_ATTR_BASE = (
    IVI_ATTR_BASE + 150000
)  # base for attributes of specific drivers


# ===== IVI Inherent Instrument Attributes ===========

# - Driver Identification

KTMAWG_ATTR_SPECIFIC_DRIVER_DESCRIPTION = (
    IVI_INHERENT_ATTR_BASE + 514
)  # ViString, read-only
# ViString, read-only
KTMAWG_ATTR_SPECIFIC_DRIVER_PREFIX = IVI_INHERENT_ATTR_BASE + 302
# ViString, read-only
KTMAWG_ATTR_SPECIFIC_DRIVER_VENDOR = IVI_INHERENT_ATTR_BASE + 513
KTMAWG_ATTR_SPECIFIC_DRIVER_REVISION = (
    IVI_INHERENT_ATTR_BASE + 551
)  # ViString, read-only
KTMAWG_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MAJOR_VERSION = (
    IVI_INHERENT_ATTR_BASE + 515
)  # ViInt32, read-only
KTMAWG_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MINOR_VERSION = (
    IVI_INHERENT_ATTR_BASE + 516
)  # ViInt32, read-only

# - User Options
# ViBoolean, read-write
KTMAWG_ATTR_RANGE_CHECK = IVI_INHERENT_ATTR_BASE + 2
KTMAWG_ATTR_QUERY_INSTRUMENT_STATUS = (
    IVI_INHERENT_ATTR_BASE + 3
)  # ViBoolean, read-write
# ViBoolean, read-write
KTMAWG_ATTR_CACHE = IVI_INHERENT_ATTR_BASE + 4
# ViBoolean, read-write
KTMAWG_ATTR_SIMULATE = IVI_INHERENT_ATTR_BASE + 5
# ViBoolean, read-write
KTMAWG_ATTR_RECORD_COERCIONS = IVI_INHERENT_ATTR_BASE + 6
# ViBoolean, read-write
KTMAWG_ATTR_INTERCHANGE_CHECK = IVI_INHERENT_ATTR_BASE + 21

# - Advanced Session Information
# ViString, read-only
KTMAWG_ATTR_LOGICAL_NAME = IVI_INHERENT_ATTR_BASE + 305
# ViString, read-only
KTMAWG_ATTR_IO_RESOURCE_DESCRIPTOR = IVI_INHERENT_ATTR_BASE + 304
# ViString, read-only
KTMAWG_ATTR_DRIVER_SETUP = IVI_INHERENT_ATTR_BASE + 7

# - Driver Capabilities
# ViString, read-only
KTMAWG_ATTR_GROUP_CAPABILITIES = IVI_INHERENT_ATTR_BASE + 401
KTMAWG_ATTR_SUPPORTED_INSTRUMENT_MODELS = (
    IVI_INHERENT_ATTR_BASE + 327
)  # ViString, read-only

# - Instrument Identification

KTMAWG_ATTR_INSTRUMENT_FIRMWARE_REVISION = (
    IVI_INHERENT_ATTR_BASE + 510
)  # ViString, read-only
KTMAWG_ATTR_INSTRUMENT_MANUFACTURER = (
    IVI_INHERENT_ATTR_BASE + 511
)  # ViString, read-only
# ViString, read-only
KTMAWG_ATTR_INSTRUMENT_MODEL = IVI_INHERENT_ATTR_BASE + 512


# ===== Instrument-Specific Attributes =====================================

# - Instrument Specific

KTMAWG_ATTR_GENERATION_SETTINGS_RESTRICTIONS = (
    IVI_SPECIFIC_ATTR_BASE + 101
)  # ViBoolean, read-only

# - System
# ViString, read-only
KTMAWG_ATTR_SERIAL_NUMBER = IVI_SPECIFIC_ATTR_BASE + 3
# ViString, read-only
KTMAWG_ATTR_SYSTEM_ABOUT = IVI_SPECIFIC_ATTR_BASE + 4
KTMAWG_ATTR_SYSTEM_GC_TIMING_OPTIMIZATION_ENABLED = (
    IVI_SPECIFIC_ATTR_BASE + 5
)  # ViBoolean, read-write
KTMAWG_ATTR_SYSTEM_IDENTIFY_ENABLED = (
    IVI_SPECIFIC_ATTR_BASE + 6
)  # ViBoolean, read-write
# ViInt32, read-only
KTMAWG_ATTR_SYSTEM_INSTANCE_ID = IVI_SPECIFIC_ATTR_BASE + 7
# ViString, read-only
KTMAWG_ATTR_SYSTEM_OPTIONS = IVI_SPECIFIC_ATTR_BASE + 8
# ViInt32, read-write
KTMAWG_ATTR_SYSTEM_TIMEOUT = IVI_SPECIFIC_ATTR_BASE + 9

# - Licensing

KTMAWG_ATTR_LICENSING_HOST_IDENTIFIER = (
    IVI_SPECIFIC_ATTR_BASE + 11
)  # ViString, read-only
KTMAWG_ATTR_LICENSING_INSTALLED_LICENSES = (
    IVI_SPECIFIC_ATTR_BASE + 12
)  # ViString, read-only

# - Calibration

KTMAWG_ATTR_CALIBRATION_ADJUSTMENT_INFORMATION = (
    IVI_SPECIFIC_ATTR_BASE + 16
)  # ViString, read-only
# ViString, read-only
KTMAWG_ATTR_CALIBRATION_DUE_DATE = IVI_SPECIFIC_ATTR_BASE + 17
KTMAWG_ATTR_CALIBRATION_INSTRUMENT_IDENTIFIER = (
    IVI_SPECIFIC_ATTR_BASE + 18
)  # ViString, read-only
# ViInt32, read-only
KTMAWG_ATTR_CALIBRATION_STATUS = IVI_SPECIFIC_ATTR_BASE + 19
KTMAWG_ATTR_CALIBRATION_VERIFICATION_INFORMATION = (
    IVI_SPECIFIC_ATTR_BASE + 20
)  # ViString, read-only

# - Module
# ViInt32, read-only
KTMAWG_ATTR_MODULE_COUNT = IVI_SPECIFIC_ATTR_BASE + 23
KTMAWG_ATTR_MODULE_INSTRUMENT_CAPABILITY = (
    IVI_SPECIFIC_ATTR_BASE + 24
)  # ViString, read-only
KTMAWG_ATTR_MODULE_MAXIMUM_RECORDED_TEMPERATURE = (
    IVI_SPECIFIC_ATTR_BASE + 25
)  # ViInt32, read-only
# ViString, read-only
KTMAWG_ATTR_MODULE_OPTIONS = IVI_SPECIFIC_ATTR_BASE + 26
# ViString, read-only
KTMAWG_ATTR_MODULE_SERIAL_NUMBER = IVI_SPECIFIC_ATTR_BASE + 27
# ViInt32, read-only
KTMAWG_ATTR_MODULE_SLOT = IVI_SPECIFIC_ATTR_BASE + 28
# ViReal64, read-only
KTMAWG_ATTR_MODULE_TEMPERATURE = IVI_SPECIFIC_ATTR_BASE + 29

# - Calibration

KTMAWG_ATTR_MODULE_CALIBRATION_ADJUSTMENT_INFORMATION = (
    IVI_SPECIFIC_ATTR_BASE + 30
)  # ViString, read-only
KTMAWG_ATTR_MODULE_CALIBRATION_DUE_DATE = (
    IVI_SPECIFIC_ATTR_BASE + 31
)  # ViString, read-only
KTMAWG_ATTR_MODULE_CALIBRATION_STATUS = (
    IVI_SPECIFIC_ATTR_BASE + 32
)  # ViInt32, read-only
KTMAWG_ATTR_MODULE_CALIBRATION_VERIFICATION_INFORMATION = (
    IVI_SPECIFIC_ATTR_BASE + 33
)  # ViString, read-only

# - Nonvolatile

KTMAWG_ATTR_NONVOLATILE_ASSET_NUMBER = (
    IVI_SPECIFIC_ATTR_BASE + 34
)  # ViString, read-write
KTMAWG_ATTR_NONVOLATILE_CAL_DUE_REMINDER = (
    IVI_SPECIFIC_ATTR_BASE + 35
)  # ViInt32, read-write
KTMAWG_ATTR_NONVOLATILE_ENABLE_INSTRUMENT_CAL_WARNINGS = (
    IVI_SPECIFIC_ATTR_BASE + 36
)  # ViBoolean, read-write
KTMAWG_ATTR_NONVOLATILE_ENABLE_MODULE_CAL_WARNINGS = (
    IVI_SPECIFIC_ATTR_BASE + 37
)  # ViBoolean, read-write
KTMAWG_ATTR_NONVOLATILE_ENABLE_PERIODIC_CAL = (
    IVI_SPECIFIC_ATTR_BASE + 38
)  # ViBoolean, read-write
KTMAWG_ATTR_NONVOLATILE_INSTRUMENT_CAL_INTERVAL = (
    IVI_SPECIFIC_ATTR_BASE + 39
)  # ViInt32, read-write
KTMAWG_ATTR_NONVOLATILE_MODULE_CAL_INTERVAL = (
    IVI_SPECIFIC_ATTR_BASE + 40
)  # ViInt32, read-write
# ViString, read-write
KTMAWG_ATTR_NONVOLATILE_PASSPHRASE = IVI_SPECIFIC_ATTR_BASE + 41
KTMAWG_ATTR_NONVOLATILE_SYSTEM_IDENTIFICATION = (
    IVI_SPECIFIC_ATTR_BASE + 42
)  # ViString, read-write

# - AssetChannel
# ViInt32, read-only
KTMAWG_ATTR_ASSETCHANNEL_COUNT = IVI_SPECIFIC_ATTR_BASE + 60

# - PeerToPeerPort
# ViInt32, read-only
KTMAWG_ATTR_PEERTOPEERPORT_COUNT = IVI_SPECIFIC_ATTR_BASE + 45
# ViString, read-write
KTMAWG_ATTR_ACTIVE_PEERTOPEERPORT = IVI_SPECIFIC_ATTR_BASE + 46

# - DeviceSync
# ViInt32, read-write
KTMAWG_ATTR_DEVICE_SYNC_CLOCK_RATE = IVI_SPECIFIC_ATTR_BASE + 88
KTMAWG_ATTR_DEVICE_SYNC_CLOCK_SOURCE = (
    IVI_SPECIFIC_ATTR_BASE + 89
)  # ViString, read-write
# ViInt32, read-write
KTMAWG_ATTR_DEVICE_SYNC_GROUP_MASK = IVI_SPECIFIC_ATTR_BASE + 90
KTMAWG_ATTR_DEVICE_SYNC_GROUP_SIGNAL = (
    IVI_SPECIFIC_ATTR_BASE + 91
)  # ViInt32, read-write
# ViInt32, read-write
KTMAWG_ATTR_DEVICE_SYNC_ROLE = IVI_SPECIFIC_ATTR_BASE + 92
KTMAWG_ATTR_DEVICE_SYNC_SLAVE_SIGNAL = (
    IVI_SPECIFIC_ATTR_BASE + 93
)  # ViInt32, read-write
# ViInt32, read-only
KTMAWG_ATTR_DEVICE_SYNC_STATE = IVI_SPECIFIC_ATTR_BASE + 94
KTMAWG_ATTR_DEVICE_SYNC_IS_ALIGNMENT_VALID = (
    IVI_SPECIFIC_ATTR_BASE + 97
)  # ViBoolean, read-only
KTMAWG_ATTR_DEVICE_SYNC_SYNC_CHANNELS = (
    IVI_SPECIFIC_ATTR_BASE + 100
)  # ViString, read-write

# - Waveform
# ViInt32, read-only
KTMAWG_ATTR_ARB_NUMBER_WAVEFORMS = IVI_SPECIFIC_ATTR_BASE + 98

# - Sequence
# ViInt32, read-only
KTMAWG_ATTR_ARB_NUMBER_SEQUENCES = IVI_SPECIFIC_ATTR_BASE + 99

# - SequenceTriggers

KTMAWG_ATTR_SEQUENCETRIGGER_MAX_COUNT = (
    IVI_SPECIFIC_ATTR_BASE + 73
)  # ViInt32, read-only
# ViInt32, read-only
KTMAWG_ATTR_SEQUENCETRIGGER_COUNT = IVI_SPECIFIC_ATTR_BASE + 74

# - Marker
# ViInt32, read-only
KTMAWG_ATTR_MARKER_COUNT = IVI_SPECIFIC_ATTR_BASE + 75
# ViInt32, read-only
KTMAWG_ATTR_MARKER_MAX_COUNT = IVI_SPECIFIC_ATTR_BASE + 76

# - External
# ViInt32, read-only
KTMAWG_ATTR_EXTERNAL_COUNT = IVI_SPECIFIC_ATTR_BASE + 79
# ViInt32, read-write
KTMAWG_ATTR_EXTERNAL_INPUT_SLOPE = IVI_SPECIFIC_ATTR_BASE + 83
KTMAWG_ATTR_EXTERNAL_OUTPUT_POLARITY = (
    IVI_SPECIFIC_ATTR_BASE + 84
)  # ViInt32, read-write
# ViReal64, read-write
KTMAWG_ATTR_EXTERNAL_THRESHOLD = IVI_SPECIFIC_ATTR_BASE + 85
# ViReal64, read-write

# - TriggerBusLine
# ViInt32, read-only
KTMAWG_ATTR_TRIGGER_BUS_LINE_COUNT = IVI_SPECIFIC_ATTR_BASE + 80
KTMAWG_ATTR_TRIGGER_BUS_LINE_INPUT_SLOPE = (
    IVI_SPECIFIC_ATTR_BASE + 81
)  # ViInt32, read-write
KTMAWG_ATTR_TRIGGER_BUS_LINE_OUTPUT_POLARITY = (
    IVI_SPECIFIC_ATTR_BASE + 82
)  # ViInt32, read-write
KTMAWG_ATTR_TRIGGER_BUS_LINE_CAPABILITY = (
    IVI_SPECIFIC_ATTR_BASE + 86
)  # ViInt32, read-only

# - ReferenceClock

KTMAWG_ATTR_REF_CLOCK_IS_PHASE_LOCKED = (
    IVI_SPECIFIC_ATTR_BASE + 96
)  # ViBoolean, read-only

# - Arbitrary
# ViReal64, read-write
KTMAWG_ATTR_ARB_GAIN = IVI_CLASS_ATTR_BASE + 202
# ViReal64, read-write
KTMAWG_ATTR_ARB_OFFSET = IVI_CLASS_ATTR_BASE + 203
# ViReal64, read-write
KTMAWG_ATTR_ARB_SAMPLE_RATE = IVI_CLASS_ATTR_BASE + 204

# - Waveform
# ViInt32, read-write
KTMAWG_ATTR_ARB_WAVEFORM_HANDLE = IVI_CLASS_ATTR_BASE + 201
# ViInt32, read-only
KTMAWG_ATTR_MAX_NUM_WAVEFORMS = IVI_CLASS_ATTR_BASE + 205
# ViInt32, read-only
KTMAWG_ATTR_WAVEFORM_QUANTUM = IVI_CLASS_ATTR_BASE + 206
# ViInt32, read-only
KTMAWG_ATTR_MAX_WAVEFORM_SIZE = IVI_CLASS_ATTR_BASE + 208
# ViInt64, read-only
KTMAWG_ATTR_MAX_WAVEFORM_SIZE64 = IVI_CLASS_ATTR_BASE + 222
# ViInt32, read-only
KTMAWG_ATTR_MIN_WAVEFORM_SIZE = IVI_CLASS_ATTR_BASE + 207
# ViInt64, read-only
KTMAWG_ATTR_MIN_WAVEFORM_SIZE64 = IVI_CLASS_ATTR_BASE + 221
# ViInt32, read-only
KTMAWG_ATTR_BINARY_ALIGNMENT = IVI_CLASS_ATTR_BASE + 241
# ViInt32, read-only
KTMAWG_ATTR_SAMPLE_BIT_RESOLUTION = IVI_CLASS_ATTR_BASE + 242
# ViInt32, read-only
KTMAWG_ATTR_OUTPUT_DATA_MASK = IVI_CLASS_ATTR_BASE + 261

# - Sequence
# ViInt32, read-write
KTMAWG_ATTR_ARB_SEQUENCE_HANDLE = IVI_CLASS_ATTR_BASE + 211
# ViInt32, read-only
KTMAWG_ATTR_MAX_SEQUENCE_LENGTH = IVI_CLASS_ATTR_BASE + 214
# ViInt32, read-only
KTMAWG_ATTR_MIN_SEQUENCE_LENGTH = IVI_CLASS_ATTR_BASE + 213
# ViInt32, read-only
KTMAWG_ATTR_MAX_LOOP_COUNT = IVI_CLASS_ATTR_BASE + 215
# ViInt32, read-only
KTMAWG_ATTR_MAX_NUM_SEQUENCES = IVI_CLASS_ATTR_BASE + 212
# ViInt32, read-only
KTMAWG_ATTR_MAX_SEQUENCE_DEPTH = IVI_CLASS_ATTR_BASE + 281

# - Output

# ViInt32, read-only
KTMAWG_ATTR_CHANNEL_COUNT = IVI_INHERENT_ATTR_BASE + 203
# ViBoolean, read-write
KTMAWG_ATTR_OUTPUT_ENABLED = IVI_CLASS_ATTR_BASE + 3
# ViReal64, read-write
KTMAWG_ATTR_OUTPUT_IMPEDANCE = IVI_CLASS_ATTR_BASE + 4
# ViInt32, read-write
KTMAWG_ATTR_OPERATION_MODE = IVI_CLASS_ATTR_BASE + 5
# ViInt32, read-write
KTMAWG_ATTR_OUTPUT_MODE = IVI_CLASS_ATTR_BASE + 1
# ViInt32, read-write
KTMAWG_ATTR_REF_CLOCK_SOURCE = IVI_CLASS_ATTR_BASE + 2
# ViInt32, read-write
KTMAWG_ATTR_TERMINAL_CONFIGURATION = IVI_CLASS_ATTR_BASE + 31

# - Trigger

# ViInt32, read-write
KTMAWG_ATTR_BURST_COUNT = IVI_CLASS_ATTR_BASE + 350
# ViInt32, read-write
KTMAWG_ATTR_TRIGGER_SOURCE = IVI_CLASS_ATTR_BASE + 302

# - Start Trigger

# ViReal64, read-write
KTMAWG_ATTR_START_TRIGGER_DELAY = IVI_CLASS_ATTR_BASE + 320
# ViInt32, read-write
KTMAWG_ATTR_START_TRIGGER_SLOPE = IVI_CLASS_ATTR_BASE + 321
# ViString, read-write
KTMAWG_ATTR_START_TRIGGER_SOURCE = IVI_CLASS_ATTR_BASE + 322
# ViReal64, read-write
KTMAWG_ATTR_START_TRIGGER_THRESHOLD = IVI_CLASS_ATTR_BASE + 323


# ***************************************************************************

# *------------------------ Attribute Value Defines -------------------------*
# ***************************************************************************

# - Defined values for

KTMAWG_VAL_STATUS_BYTE_FLAGS_USER0 = 1
KTMAWG_VAL_STATUS_BYTE_FLAGS_USER1 = 2
KTMAWG_VAL_STATUS_BYTE_FLAGS_USER2 = 4
KTMAWG_VAL_STATUS_BYTE_FLAGS_USER3 = 8
KTMAWG_VAL_STATUS_BYTE_FLAGS_MESSAGE_AVAILABLE = 16
KTMAWG_VAL_STATUS_BYTE_FLAGS_EVENT_STATUS_REGISTER = 32
KTMAWG_VAL_STATUS_BYTE_FLAGS_REQUESTING_SERVICE = 64
KTMAWG_VAL_STATUS_BYTE_FLAGS_USER7 = 128

# - Defined values for

KTMAWG_VAL_UPDATE_STATUS_IN_PROGRESS = 0
KTMAWG_VAL_UPDATE_STATUS_SUCCESS = 1
KTMAWG_VAL_UPDATE_STATUS_RESTART_REQUIRED = 2
KTMAWG_VAL_UPDATE_STATUS_POWER_CYCLE_REQUIRED = 3
KTMAWG_VAL_UPDATE_STATUS_FAILED = 4

# - Defined values for
#  attribute KTMAWG_ATTR_CALIBRATION_STATUS
#  attribute KTMAWG_ATTR_MODULE_CALIBRATION_STATUS

KTMAWG_VAL_CALIBRATION_STATUS_DUE = 1
KTMAWG_VAL_CALIBRATION_STATUS_EXPIRED = 2
KTMAWG_VAL_CALIBRATION_STATUS_INSTRUMENT_CALIBRATED = 0
KTMAWG_VAL_CALIBRATION_STATUS_MODULES_CALIBRATED = 5
KTMAWG_VAL_CALIBRATION_STATUS_NOT_CALIBRATED = 3
KTMAWG_VAL_CALIBRATION_STATUS_NOT_SUBJECT_TO_CALIBRATION = 4

# - Defined values for
#   parameter Mask in function KtMAwg_StatusClearEvent
#   parameter Mask in function KtMAwg_StatusGetEventState
#   parameter Mask in function KtMAwg_StatusSetEvent

KTMAWG_VAL_STATUS_EVENT_ALL = 0
KTMAWG_VAL_STATUS_EVENT_ERROR_IN_QUEUE = 1
KTMAWG_VAL_STATUS_EVENT_THERMAL_SHUTDOWN = 14
KTMAWG_VAL_STATUS_EVENT_THERMAL_WARNING = 15
KTMAWG_VAL_STATUS_EVENT_USER_DEFINED = 2

# - Defined values for

KTMAWG_VAL_MODULE_EVENT_ALL = 0
KTMAWG_VAL_MODULE_EVENT_TIMER1 = 201
KTMAWG_VAL_MODULE_EVENT_TIMER2 = 202
KTMAWG_VAL_MODULE_EVENT_TIMER3 = 203
KTMAWG_VAL_MODULE_EVENT_TIMER4 = 204
KTMAWG_VAL_MODULE_EVENT_TRIGGER1 = 1
KTMAWG_VAL_MODULE_EVENT_TRIGGER10 = 10
KTMAWG_VAL_MODULE_EVENT_TRIGGER2 = 2
KTMAWG_VAL_MODULE_EVENT_TRIGGER3 = 3
KTMAWG_VAL_MODULE_EVENT_TRIGGER4 = 4
KTMAWG_VAL_MODULE_EVENT_TRIGGER5 = 5
KTMAWG_VAL_MODULE_EVENT_TRIGGER6 = 6
KTMAWG_VAL_MODULE_EVENT_TRIGGER7 = 7
KTMAWG_VAL_MODULE_EVENT_TRIGGER8 = 8
KTMAWG_VAL_MODULE_EVENT_TRIGGER9 = 9
KTMAWG_VAL_MODULE_EVENT_USER1 = 101
KTMAWG_VAL_MODULE_EVENT_USER2 = 102
KTMAWG_VAL_MODULE_EVENT_USER3 = 103
KTMAWG_VAL_MODULE_EVENT_USER4 = 104

# - Defined values for

KTMAWG_VAL_SPI_MODE_LSB_FIRST_BYTE = 0
KTMAWG_VAL_SPI_MODE_LSB_FIRST_WORD = 1
KTMAWG_VAL_SPI_MODE_MSB_FIRST_BYTE = 2
KTMAWG_VAL_SPI_MODE_MSB_FIRST_WORD = 3

# - Defined values for

KTMAWG_VAL_TRIGGER_DIRECTION_INPUT = 0
KTMAWG_VAL_TRIGGER_DIRECTION_OUTPUT = 1

# - Defined values for

KTMAWG_VAL_TRIGGER_MODE_LEVEL = 1
KTMAWG_VAL_TRIGGER_MODE_PULSE = 0

# - Defined values for

KTMAWG_VAL_TRIGGER_POLARITY_NEGATIVE = 1
KTMAWG_VAL_TRIGGER_POLARITY_POSITIVE = 0

# - Defined values for

KTMAWG_VAL_TRIGGER_PXI_TRIGGER0 = 0
KTMAWG_VAL_TRIGGER_PXI_TRIGGER1 = 1
KTMAWG_VAL_TRIGGER_PXI_TRIGGER2 = 2
KTMAWG_VAL_TRIGGER_PXI_TRIGGER3 = 3
KTMAWG_VAL_TRIGGER_PXI_TRIGGER4 = 4
KTMAWG_VAL_TRIGGER_PXI_TRIGGER5 = 5
KTMAWG_VAL_TRIGGER_PXI_TRIGGER6 = 6
KTMAWG_VAL_TRIGGER_PXI_TRIGGER7 = 7

# - Defined values for

KTMAWG_VAL_TRIGGER_RESOURCE_NONE = -1
KTMAWG_VAL_TRIGGER_RESOURCE_TRIGGER1 = 0
KTMAWG_VAL_TRIGGER_RESOURCE_TRIGGER2 = 1
KTMAWG_VAL_TRIGGER_RESOURCE_TRIGGER3 = 2
KTMAWG_VAL_TRIGGER_RESOURCE_TRIGGER4 = 3
KTMAWG_VAL_TRIGGER_RESOURCE_TRIGGER5 = 4
KTMAWG_VAL_TRIGGER_RESOURCE_TRIGGER6 = 5
KTMAWG_VAL_TRIGGER_RESOURCE_TRIGGER7 = 6
KTMAWG_VAL_TRIGGER_RESOURCE_TRIGGER8 = 7
KTMAWG_VAL_TRIGGER_RESOURCE_TRIGGER9 = 8
KTMAWG_VAL_TRIGGER_RESOURCE_TRIGGER10 = 9

# - Defined values for

KTMAWG_VAL_TRIGGER_TERMINATION_HIGH_IMPEDANCE = 0
KTMAWG_VAL_TRIGGER_TERMINATION_LOW_IMPEDANCE = 1

# - Defined values for

KTMAWG_VAL_LTC3880_PHASE_BOTH_PHASES = 255
KTMAWG_VAL_LTC3880_PHASE_PHASE0 = 0
KTMAWG_VAL_LTC3880_PHASE_PHASE1 = 1

# - Defined values for

KTMAWG_VAL_SELF_TEST_ADC_MEASUREMENT_MODE_COMMON_REFERENCE = 1
KTMAWG_VAL_SELF_TEST_ADC_MEASUREMENT_MODE_DIFFERENTIAL = 0

# - Defined values for

KTMAWG_VAL_SELF_TEST_ADC_MEASUREMENT_RANGE_VOLTAGE_0P64 = 5
KTMAWG_VAL_SELF_TEST_ADC_MEASUREMENT_RANGE_VOLTAGE1_0P24 = 1
KTMAWG_VAL_SELF_TEST_ADC_MEASUREMENT_RANGE_VOLTAGE_1P28 = 4
KTMAWG_VAL_SELF_TEST_ADC_MEASUREMENT_RANGE_VOLTAGE2_0P48 = 7
KTMAWG_VAL_SELF_TEST_ADC_MEASUREMENT_RANGE_VOLTAGE2_4P576 = 0
KTMAWG_VAL_SELF_TEST_ADC_MEASUREMENT_RANGE_VOLTAGE_2P56 = 3
KTMAWG_VAL_SELF_TEST_ADC_MEASUREMENT_RANGE_VOLTAGE_5P12 = 2

# - Defined values for
#   parameter State in function KtMAwg_GetGenerationStateByChannel

KTMAWG_VAL_GENERATION_STATE_CONFIGURATION = 0
KTMAWG_VAL_GENERATION_STATE_OUTPUT_GENERATION = 1

# - Defined values for

KTMAWG_VAL_MEMORY_QUERY_MODE_TOTAL_SIZE = 0
KTMAWG_VAL_MEMORY_QUERY_MODE_FREE_SIZE = 1
KTMAWG_VAL_MEMORY_QUERY_MODE_ALLOCATED_SIZE = 2

# - Defined values for
#   parameter Mode in function KtMAwg_CorrectionsGetFilterMode
#   parameter Mode in function KtMAwg_CorrectionsSetFilterMode

KTMAWG_VAL_CORRECTION_FILTER_MODE_UNCALIBRATED = 1
KTMAWG_VAL_CORRECTION_FILTER_MODE_BYPASS = 2
KTMAWG_VAL_CORRECTION_FILTER_MODE_CALIBRATED = 0

# - Defined values for

KTMAWG_VAL_ATTRIBUTES_ARBITRARY_SAMPLE_RATE = 0
KTMAWG_VAL_ATTRIBUTES_OUTPUT_DELAY = 4
KTMAWG_VAL_ATTRIBUTES_MARKER_DELAY = 20
KTMAWG_VAL_ATTRIBUTES_ARBITRARY_GAIN = 10
KTMAWG_VAL_ATTRIBUTES_ARBITRARY_ANALOG_GAIN = 13
KTMAWG_VAL_ATTRIBUTES_ARBITRARY_DIGITAL_GAIN = 14
KTMAWG_VAL_ATTRIBUTES_ARBITRARY_OFFSET = 11
KTMAWG_VAL_ATTRIBUTES_ARBITRARY_COMMON_MODE_OFFSET = 12
KTMAWG_VAL_ATTRIBUTES_ARBITRARY_LOAD_IMPEDANCE = 8
KTMAWG_VAL_ATTRIBUTES_START_TRIGGER_DELAY = 16
KTMAWG_VAL_ATTRIBUTES_START_TRIGGER_SOURCE = 15
KTMAWG_VAL_ATTRIBUTES_MARKER_DESTINATION = 19
KTMAWG_VAL_ATTRIBUTES_MARKER_BIT_POSITION = 21
KTMAWG_VAL_ATTRIBUTES_SEQUENCE_TRIGGER_SOURCE = 17
KTMAWG_VAL_ATTRIBUTES_SEQUENCE_TRIGGER_DELAY = 18
KTMAWG_VAL_ATTRIBUTES_OUTPUT_OPERATION_MODE = 5
KTMAWG_VAL_ATTRIBUTES_ARBITRARY_GAIN_CONTROL = 9
KTMAWG_VAL_ATTRIBUTES_ARBITRARY_CHANNEL_MODE = 2
KTMAWG_VAL_ATTRIBUTES_ARBITRARY_HANDLE = 1
KTMAWG_VAL_ATTRIBUTES_OUTPUT_ENABLED = 3
KTMAWG_VAL_ATTRIBUTES_OUTPUT_TERMINAL_CONFIGURATION = 7
KTMAWG_VAL_ATTRIBUTES_CORRECTION_FILTER_MODE = 22
KTMAWG_VAL_ATTRIBUTES_BURST_COUNT = 6
KTMAWG_VAL_ATTRIBUTES_ARBITRARY_WAVEFORM_MODE = 30

# - Defined values for
#   parameter Control in function KtMAwg_ArbitrarySetGainControl
#   parameter Control in function KtMAwg_ArbitraryGetGainControl

KTMAWG_VAL_GAIN_CONTROL_COMPONENT = 1
KTMAWG_VAL_GAIN_CONTROL_COMPOSITE = 0

# - Defined values for
#   parameter Events in function KtMAwg_QueryGenerationEvents

KTMAWG_VAL_GENERATION_EVENTS_NONE = 0
KTMAWG_VAL_GENERATION_EVENTS_DIGITAL_GAIN_BLOCK_OVERFLOW = 1
KTMAWG_VAL_GENERATION_EVENTS_FIFO_UNDERRUN = 2
KTMAWG_VAL_GENERATION_EVENTS_RESAMPLER_OVERFLOW = 4
KTMAWG_VAL_GENERATION_EVENTS_CORRECTIONS_OVERFLOW = 8
KTMAWG_VAL_GENERATION_EVENTS_MARKER_QUEUE_OVERRUN = 16
KTMAWG_VAL_GENERATION_EVENTS_MARKER_IGNORED = 32
KTMAWG_VAL_GENERATION_EVENTS_STREAMING_DATA_UNDERRUN = 128

# - Defined values for
#   attribute KTMAWG_ATTR_OUTPUT_MODE
#   parameter OutputMode in function KtMAwg_ConfigureOutputMode

KTMAWG_VAL_OUTPUT_FUNC = 0
KTMAWG_VAL_OUTPUT_ARB = 1
KTMAWG_VAL_OUTPUT_SEQ = 2

# - Defined values for
#   attribute KTMAWG_ATTR_OPERATION_MODE
#   parameter OperationMode in function KtMAwg_ConfigureOperationMode

KTMAWG_VAL_OPERATE_CONTINUOUS = 0
KTMAWG_VAL_OPERATE_BURST = 1

# - Defined values for
#   attribute KTMAWG_ATTR_REF_CLOCK_SOURCE
#   parameter Source in function KtMAwg_ConfigureRefClockSource

KTMAWG_VAL_REF_CLOCK_INTERNAL = 0
KTMAWG_VAL_REF_CLOCK_EXTERNAL = 1
KTMAWG_VAL_REF_CLOCK_RTSI_CLOCK = 101

# - Defined values for
#   attribute KTMAWG_ATTR_TRIGGER_SOURCE
#   parameter Source in function KtMAwg_ConfigureTriggerSource

KTMAWG_VAL_EXTERNAL = 1
KTMAWG_VAL_SOFTWARE_TRIG = 2
KTMAWG_VAL_INTERNAL_TRIGGER = 3
KTMAWG_VAL_TTL0 = 111
KTMAWG_VAL_TTL1 = 112
KTMAWG_VAL_TTL2 = 113
KTMAWG_VAL_TTL3 = 114
KTMAWG_VAL_TTL4 = 115
KTMAWG_VAL_TTL5 = 116
KTMAWG_VAL_TTL6 = 117
KTMAWG_VAL_TTL7 = 118
KTMAWG_VAL_ECL0 = 119
KTMAWG_VAL_ECL1 = 120
KTMAWG_VAL_PXI_STAR = 131
KTMAWG_VAL_RTSI_0 = 141
KTMAWG_VAL_RTSI_1 = 142
KTMAWG_VAL_RTSI_2 = 143
KTMAWG_VAL_RTSI_3 = 144
KTMAWG_VAL_RTSI_4 = 145
KTMAWG_VAL_RTSI_5 = 146
KTMAWG_VAL_RTSI_6 = 147
KTMAWG_VAL_SOFTWARE0 = 1001
KTMAWG_VAL_SOFTWARE1 = 1002
KTMAWG_VAL_SOFTWARE2 = 1003
KTMAWG_VAL_SOFTWARE3 = 1004
KTMAWG_VAL_SOFTWARE4 = 1005
KTMAWG_VAL_SOFTWARE5 = 1006
KTMAWG_VAL_SOFTWARE6 = 1007
KTMAWG_VAL_SOFTWARE7 = 1008
KTMAWG_VAL_SOFTWARE9 = 1010
KTMAWG_VAL_SOFTWARE10 = 1011
KTMAWG_VAL_SOFTWARE11 = 1012
KTMAWG_VAL_SOFTWARE12 = 1013
KTMAWG_VAL_SOFTWARE13 = 1014
KTMAWG_VAL_SOFTWARE14 = 1015
KTMAWG_VAL_SOFTWARE15 = 1016
KTMAWG_VAL_SOFTWARE8 = 1009
KTMAWG_VAL_PXI_TRIG0 = 1017
KTMAWG_VAL_PXI_TRIG1 = 1018
KTMAWG_VAL_PXI_TRIG2 = 1019
KTMAWG_VAL_PXI_TRIG3 = 1020
KTMAWG_VAL_PXI_TRIG4 = 1021
KTMAWG_VAL_PXI_TRIG5 = 1022
KTMAWG_VAL_PXI_TRIG6 = 1023
KTMAWG_VAL_PXI_TRIG7 = 1024
KTMAWG_VAL_EXTERNAL1 = 1025
KTMAWG_VAL_EXTERNAL2 = 1026
KTMAWG_VAL_PXIE_DSTARB = 1027
KTMAWG_VAL_DIO_0 = 1028
KTMAWG_VAL_DIO_1 = 1029
KTMAWG_VAL_DIO_2 = 1030
KTMAWG_VAL_DIO_3 = 1031
KTMAWG_VAL_DIO_4 = 1032
KTMAWG_VAL_DIO_5 = 1033
KTMAWG_VAL_DIO_7 = 1035
KTMAWG_VAL_DIO_6 = 1034
KTMAWG_VAL_TRIG_IMMEDIATE = 1036
KTMAWG_VAL_DISABLED = 1037

# - Defined values for

KTMAWG_VAL_WFM_SINE = 1
KTMAWG_VAL_WFM_SQUARE = 2
KTMAWG_VAL_WFM_TRIANGLE = 3
KTMAWG_VAL_WFM_RAMP_UP = 4
KTMAWG_VAL_WFM_RAMP_DOWN = 5
KTMAWG_VAL_WFM_DC = 6

# - Defined values for

KTMAWG_VAL_AM_INTERNAL_SINE = 1
KTMAWG_VAL_AM_INTERNAL_SQUARE = 2
KTMAWG_VAL_AM_INTERNAL_TRIANGLE = 3
KTMAWG_VAL_AM_INTERNAL_RAMP_UP = 4
KTMAWG_VAL_AM_INTERNAL_RAMP_DOWN = 5

# - Defined values for

KTMAWG_VAL_AM_INTERNAL = 0
KTMAWG_VAL_AM_EXTERNAL = 1

# - Defined values for

KTMAWG_VAL_FM_INTERNAL_SINE = 1
KTMAWG_VAL_FM_INTERNAL_SQUARE = 2
KTMAWG_VAL_FM_INTERNAL_TRIANGLE = 3
KTMAWG_VAL_FM_INTERNAL_RAMP_UP = 4
KTMAWG_VAL_FM_INTERNAL_RAMP_DOWN = 5

# - Defined values for

KTMAWG_VAL_FM_INTERNAL = 0
KTMAWG_VAL_FM_EXTERNAL = 1

# - Defined values for
#   attribute KTMAWG_ATTR_START_TRIGGER_SLOPE
#   parameter Slope in function KtMAwg_ConfigureStartTrigger

KTMAWG_VAL_TRIGGER_POSITIVE = 0
KTMAWG_VAL_TRIGGER_NEGATIVE = 1
KTMAWG_VAL_TRIGGER_EITHER = 2

# - Defined values for
#   attribute KTMAWG_ATTR_TERMINAL_CONFIGURATION
#   parameter Configuration in function KtMAwg_OutputSetTerminalConfiguration

KTMAWG_VAL_TERMINAL_CONFIGURATION_SINGLE_ENDED = 0
KTMAWG_VAL_TERMINAL_CONFIGURATION_DIFFERENTIAL = 1

# - Defined values for

KTMAWG_VAL_SAMPLE_CLOCK_SOURCE_INTERNAL = 0
KTMAWG_VAL_SAMPLE_CLOCK_SOURCE_EXTERNAL = 1

# - Defined values for

KTMAWG_VAL_MARKER_POLARITY_ACTIVE_HIGH = 0
KTMAWG_VAL_MARKER_POLARITY_ACTIVE_LOW = 1

# - Defined values for
#   attribute KTMAWG_ATTR_BINARY_ALIGNMENT

KTMAWG_VAL_BINARY_ALIGNMENT_LEFT = 0
KTMAWG_VAL_BINARY_ALIGNMENT_RIGHT = 1

# - Defined values for
#   parameter Mode in function KtMAwg_GetChannelMode
#   parameter Mode in function KtMAwg_SetChannelMode

KTMAWG_VAL_CHANNEL_MODE_WAVEFORM = 0
KTMAWG_VAL_CHANNEL_MODE_MARKER = 1

# - Defined values for

KTMAWG_VAL_EXTERNAL_IMPEDANCE_OHM_1K = 0
KTMAWG_VAL_EXTERNAL_IMPEDANCE_OHM50 = 1

# - Defined values for

KTMAWG_VAL_EXTERNAL_ATTRIBUTES_THRESHOLD = 0

# - Defined values for
#   attribute KTMAWG_ATTR_EXTERNAL_INPUT_SLOPE
#   attribute KTMAWG_ATTR_TRIGGER_BUS_LINE_INPUT_SLOPE

KTMAWG_VAL_INPUT_SLOPE_POSITIVE = 0
KTMAWG_VAL_INPUT_SLOPE_NEGATIVE = 1

# - Defined values for
#   attribute KTMAWG_ATTR_EXTERNAL_OUTPUT_POLARITY
#   attribute KTMAWG_ATTR_TRIGGER_BUS_LINE_OUTPUT_POLARITY
#   parameter Val in function KtMAwg_GetChannelMarkerPolarity
#  parameter Polarity in function KtMAwg_SetChannelMarkerPolarity

KTMAWG_VAL_OUTPUT_POLARITY_POSITIVE = 0
KTMAWG_VAL_OUTPUT_POLARITY_NEGATIVE = 1

# - Defined values for
#   parameter Option in function KtMAwg_MemoryTest

KTMAWG_VAL_MEMORY_TEST_STANDARD = 1
KTMAWG_VAL_MEMORY_TEST_COMPREHENSIVE = 0

# - Defined values for
#   attribute KTMAWG_ATTR_TRIGGER_BUS_LINE_CAPABILITY

KTMAWG_VAL_TRIGGER_BUS_LINE_CAPABILITY_INPUT = 1
KTMAWG_VAL_TRIGGER_BUS_LINE_CAPABILITY_OUTPUT = 2
KTMAWG_VAL_TRIGGER_BUS_LINE_CAPABILITY_BOTH = 3

# - Defined values for
#   attribute KTMAWG_ATTR_DEVICE_SYNC_GROUP_SIGNAL
#   attribute KTMAWG_ATTR_DEVICE_SYNC_SLAVE_SIGNAL
#   parameter GroupSignal in function KtMAwg_DeviceSyncConfigureMaster
#   parameter SlaveSignals in function KtMAwg_DeviceSyncConfigureMaster
#   parameter GroupSignal in function KtMAwg_DeviceSyncConfigureSlave
#   parameter SlaveSignal in function KtMAwg_DeviceSyncConfigureSlave
#   parameter Val in function KtMAwg_DeviceSyncGetGroupMask
#   parameter SlaveSignals in function KtMAwg_DeviceSyncSetGroupMask

KTMAWG_VAL_DEVICE_SYNC_RESOURCESFP_SYNC = 32768
KTMAWG_VAL_DEVICE_SYNC_RESOURCES_NONE = 0
KTMAWG_VAL_DEVICE_SYNC_RESOURCESPXI_LBL6 = 2048
KTMAWG_VAL_DEVICE_SYNC_RESOURCESPXI_LBR6 = 1024
KTMAWG_VAL_DEVICE_SYNC_RESOURCESPXI_STAR = 512
KTMAWG_VAL_DEVICE_SYNC_RESOURCESPXI_TRIG0 = 1
KTMAWG_VAL_DEVICE_SYNC_RESOURCESPXI_TRIG1 = 2
KTMAWG_VAL_DEVICE_SYNC_RESOURCESPXI_TRIG2 = 4
KTMAWG_VAL_DEVICE_SYNC_RESOURCESPXI_TRIG3 = 8
KTMAWG_VAL_DEVICE_SYNC_RESOURCESPXI_TRIG4 = 16
KTMAWG_VAL_DEVICE_SYNC_RESOURCESPXI_TRIG5 = 32
KTMAWG_VAL_DEVICE_SYNC_RESOURCESPXI_TRIG6 = 64
KTMAWG_VAL_DEVICE_SYNC_RESOURCESPXI_TRIG7 = 128
KTMAWG_VAL_DEVICE_SYNC_RESOURCESPXIE_DSTARA = 4096
KTMAWG_VAL_DEVICE_SYNC_RESOURCESPXIE_DSTARB = 8192
KTMAWG_VAL_DEVICE_SYNC_RESOURCESPXIE_DSTARC = 16384

# - Defined values for
#   attribute KTMAWG_ATTR_DEVICE_SYNC_ROLE
#   parameter Role in function KtMAwg_DeviceSyncConfigureMaster

KTMAWG_VAL_DEVICE_SYNC_ROLE_GROUP_MASTER = 3
KTMAWG_VAL_DEVICE_SYNC_ROLE_LOCAL_MASTER = 5
KTMAWG_VAL_DEVICE_SYNC_ROLE_OFF = 0
KTMAWG_VAL_DEVICE_SYNC_ROLE_SLAVE = 2
KTMAWG_VAL_DEVICE_SYNC_ROLE_SANDBOX = 1
KTMAWG_VAL_DEVICE_SYNC_ROLE_SYSTEM_MASTER = 4

# - Defined values for
#   attribute KTMAWG_ATTR_DEVICE_SYNC_STATE

KTMAWG_VAL_DEVICE_SYNC_STATE_ARM = 1
KTMAWG_VAL_DEVICE_SYNC_STATE_IDLE = 0
KTMAWG_VAL_DEVICE_SYNC_STATE_RUN = 3
KTMAWG_VAL_DEVICE_SYNC_STATE_TRIGGER = 2
KTMAWG_VAL_DEVICE_SYNC_STATE_UNKNOWN = 4

# - Defined values for
#   parameter State in function KtMAwg_GetOutputGenerationState

KTMAWG_VAL_OUTPUT_GENERATION_STATE_IDLE = 0
KTMAWG_VAL_OUTPUT_GENERATION_STATE_ARM = 1
KTMAWG_VAL_OUTPUT_GENERATION_STATE_TRIGGER = 2
KTMAWG_VAL_OUTPUT_GENERATION_STATE_RUN = 3
KTMAWG_VAL_OUTPUT_GENERATION_STATE_UNKNOWN = 4

# - Defined values for
#   parameter Val in function KtMAwg_GetWaveformMode
#   parameter Mode in function KtMAwg_SetWaveformMode

KTMAWG_VAL_WAVEFORM_MODE_CATALOGED = 0
KTMAWG_VAL_WAVEFORM_MODE_STREAMING = 1


# ***************************************************************************
# *----------------- Instrument Error And Completion Codes ------------------*
# ***************************************************************************

IVIC_WARN_BASE = 0x3FFA0000
IVIC_CROSS_CLASS_WARN_BASE = IVIC_WARN_BASE + 0x1000
IVIC_CLASS_WARN_BASE = IVIC_WARN_BASE + 0x2000
IVIC_SPECIFIC_WARN_BASE = IVIC_WARN_BASE + 0x4000

IVIC_ERROR_BASE = 0xBFFA0000
IVIC_CROSS_CLASS_ERROR_BASE = IVIC_ERROR_BASE + 0x1000
IVIC_CLASS_ERROR_BASE = IVIC_ERROR_BASE + 0x2000
IVIC_SPECIFIC_ERROR_BASE = IVIC_ERROR_BASE + 0x4000
IVIC_LXISYNC_ERROR_BASE = IVIC_ERROR_BASE + 0x2000


KTMAWG_ERROR_CANNOT_RECOVER = IVIC_ERROR_BASE + 0x0000
KTMAWG_ERROR_INSTRUMENT_STATUS = IVIC_ERROR_BASE + 0x0001
KTMAWG_ERROR_CANNOT_OPEN_FILE = IVIC_ERROR_BASE + 0x0002
KTMAWG_ERROR_READING_FILE = IVIC_ERROR_BASE + 0x0003
KTMAWG_ERROR_WRITING_FILE = IVIC_ERROR_BASE + 0x0004
KTMAWG_ERROR_INVALID_PATHNAME = IVIC_ERROR_BASE + 0x000B
KTMAWG_ERROR_INVALID_ATTRIBUTE = IVIC_ERROR_BASE + 0x000C
KTMAWG_ERROR_IVI_ATTR_NOT_WRITABLE = IVIC_ERROR_BASE + 0x000D
KTMAWG_ERROR_IVI_ATTR_NOT_READABLE = IVIC_ERROR_BASE + 0x000E
KTMAWG_ERROR_INVALID_VALUE = IVIC_ERROR_BASE + 0x0010
KTMAWG_ERROR_FUNCTION_NOT_SUPPORTED = IVIC_ERROR_BASE + 0x0011
KTMAWG_ERROR_ATTRIBUTE_NOT_SUPPORTED = IVIC_ERROR_BASE + 0x0012
KTMAWG_ERROR_VALUE_NOT_SUPPORTED = IVIC_ERROR_BASE + 0x0013
KTMAWG_ERROR_TYPES_DO_NOT_MATCH = IVIC_ERROR_BASE + 0x0015
KTMAWG_ERROR_NOT_INITIALIZED = IVIC_ERROR_BASE + 0x001D
KTMAWG_ERROR_UNKNOWN_CHANNEL_NAME = IVIC_ERROR_BASE + 0x0020
KTMAWG_ERROR_TOO_MANY_OPEN_FILES = IVIC_ERROR_BASE + 0x0023
KTMAWG_ERROR_CHANNEL_NAME_REQUIRED = IVIC_ERROR_BASE + 0x0044
KTMAWG_ERROR_MISSING_OPTION_NAME = IVIC_ERROR_BASE + 0x0049
KTMAWG_ERROR_MISSING_OPTION_VALUE = IVIC_ERROR_BASE + 0x004A
KTMAWG_ERROR_BAD_OPTION_NAME = IVIC_ERROR_BASE + 0x004B
KTMAWG_ERROR_BAD_OPTION_VALUE = IVIC_ERROR_BASE + 0x004C
KTMAWG_ERROR_OUT_OF_MEMORY = IVIC_ERROR_BASE + 0x0056
KTMAWG_ERROR_OPERATION_PENDING = IVIC_ERROR_BASE + 0x0057
KTMAWG_ERROR_NULL_POINTER = IVIC_ERROR_BASE + 0x0058
KTMAWG_ERROR_UNEXPECTED_RESPONSE = IVIC_ERROR_BASE + 0x0059
KTMAWG_ERROR_FILE_NOT_FOUND = IVIC_ERROR_BASE + 0x005B
KTMAWG_ERROR_INVALID_FILE_FORMAT = IVIC_ERROR_BASE + 0x005C
KTMAWG_ERROR_STATUS_NOT_AVAILABLE = IVIC_ERROR_BASE + 0x005D
KTMAWG_ERROR_ID_QUERY_FAILED = IVIC_ERROR_BASE + 0x005E
KTMAWG_ERROR_RESET_FAILED = IVIC_ERROR_BASE + 0x005F
KTMAWG_ERROR_RESOURCE_UNKNOWN = IVIC_ERROR_BASE + 0x0060
KTMAWG_ERROR_ALREADY_INITIALIZED = IVIC_ERROR_BASE + 0x0061
KTMAWG_ERROR_CANNOT_CHANGE_SIMULATION_STATE = IVIC_ERROR_BASE + 0x0062
KTMAWG_ERROR_INVALID_NUMBER_OF_LEVELS_IN_SELECTOR = IVIC_ERROR_BASE + 0x0063
KTMAWG_ERROR_INVALID_RANGE_IN_SELECTOR = IVIC_ERROR_BASE + 0x0064
KTMAWG_ERROR_UNKOWN_NAME_IN_SELECTOR = IVIC_ERROR_BASE + 0x0065
KTMAWG_ERROR_BADLY_FORMED_SELECTOR = IVIC_ERROR_BASE + 0x0066
KTMAWG_ERROR_UNKNOWN_PHYSICAL_IDENTIFIER = IVIC_ERROR_BASE + 0x0067
KTMAWG_ERROR_INVALID_SESSION_HANDLE = IVIC_ERROR_BASE + 0x1190


KTMAWG_SUCCESS = 0
KTMAWG_WARN_NSUP_ID_QUERY = IVIC_WARN_BASE + 0x0065
KTMAWG_WARN_NSUP_RESET = IVIC_WARN_BASE + 0x0066
KTMAWG_WARN_NSUP_SELF_TEST = IVIC_WARN_BASE + 0x0067
KTMAWG_WARN_NSUP_ERROR_QUERY = IVIC_WARN_BASE + 0x0068
KTMAWG_WARN_NSUP_REV_QUERY = IVIC_WARN_BASE + 0x0069


KTMAWG_ERROR_IO_GENERAL = IVIC_SPECIFIC_ERROR_BASE + 0x0214
KTMAWG_ERROR_IO_TIMEOUT = IVIC_SPECIFIC_ERROR_BASE + 0x0215
KTMAWG_ERROR_CALIBRATION_VERSION = IVIC_SPECIFIC_ERROR_BASE + 0x0217
KTMAWG_ERROR_CHECK_ERROR_QUEUE = IVIC_SPECIFIC_ERROR_BASE + 0x0218
KTMAWG_ERROR_FILE_TYPE_NOT_RECOGNIZED = IVIC_SPECIFIC_ERROR_BASE + 0x0219
KTMAWG_ERROR_FIRMWARE_UPDATE_ERROR = IVIC_SPECIFIC_ERROR_BASE + 0x021A
KTMAWG_ERROR_FIRMWARE_UPDATE_REQUIRED = IVIC_SPECIFIC_ERROR_BASE + 0x021B
KTMAWG_ERROR_FPGA_PROGRAMMING_ERROR = IVIC_SPECIFIC_ERROR_BASE + 0x021C
KTMAWG_ERROR_HARDWARE_TIMEOUT = IVIC_SPECIFIC_ERROR_BASE + 0x021D
KTMAWG_ERROR_HW_RESOURCE_NOT_AVAILABLE = IVIC_SPECIFIC_ERROR_BASE + 0x021E
KTMAWG_ERROR_INCOMPATIBLE_SOFTWARE_VERSION_ERROR = \
    IVIC_SPECIFIC_ERROR_BASE + 0x021F
KTMAWG_ERROR_INSTRUMENT_CALIBRATION_DUE = IVIC_SPECIFIC_ERROR_BASE + 0x0220
KTMAWG_ERROR_INSTRUMENT_CALIBRATION_EXPIRED = IVIC_SPECIFIC_ERROR_BASE + 0x0221
KTMAWG_ERROR_INSTRUMENT_NOT_CALIBRATED = IVIC_SPECIFIC_ERROR_BASE + 0x0222
KTMAWG_ERROR_INTERNAL_APPLICATION_ERROR = IVIC_SPECIFIC_ERROR_BASE + 0x0223
KTMAWG_ERROR_LICENSE_SYSTEM_ERROR = IVIC_SPECIFIC_ERROR_BASE + 0x0224
KTMAWG_ERROR_LICENSE_VALIDATION_ERROR = IVIC_SPECIFIC_ERROR_BASE + 0x0225
KTMAWG_ERROR_MAX_TIME_EXCEEDED = IVIC_SPECIFIC_ERROR_BASE + 0x0226
KTMAWG_ERROR_MISSING_CLOCK_ERROR = IVIC_SPECIFIC_ERROR_BASE + 0x0227
KTMAWG_ERROR_MODEL_NOT_SUPPORTED = IVIC_SPECIFIC_ERROR_BASE + 0x0228
KTMAWG_ERROR_MODULE_CALIBRATION_DUE = IVIC_SPECIFIC_ERROR_BASE + 0x0229
KTMAWG_ERROR_MODULE_CALIBRATION_EXPIRED = IVIC_SPECIFIC_ERROR_BASE + 0x022A
KTMAWG_ERROR_MODULE_NOT_CALIBRATED = IVIC_SPECIFIC_ERROR_BASE + 0x022B
KTMAWG_ERROR_MODULE_VALIDATE_FAILED = IVIC_SPECIFIC_ERROR_BASE + 0x022C
KTMAWG_ERROR_OPERATION_ABORTED = IVIC_SPECIFIC_ERROR_BASE + 0x022D
KTMAWG_ERROR_OPERATION_ERROR = IVIC_SPECIFIC_ERROR_BASE + 0x022E
KTMAWG_ERROR_OUT_OF_MEMORY_ERROR = IVIC_SPECIFIC_ERROR_BASE + 0x022F
KTMAWG_ERROR_PARAMETER_VALIDATION_ERROR = IVIC_SPECIFIC_ERROR_BASE + 0x0230
KTMAWG_ERROR_PERSONALITY_NOT_ACTIVE = IVIC_SPECIFIC_ERROR_BASE + 0x0231
KTMAWG_ERROR_PERSONALITY_NOT_INSTALLED = IVIC_SPECIFIC_ERROR_BASE + 0x0232
KTMAWG_ERROR_PERSONALITY_NOT_LICENSED = IVIC_SPECIFIC_ERROR_BASE + 0x0233
KTMAWG_ERROR_SELFTEST_FAILED = IVIC_SPECIFIC_ERROR_BASE + 0x0234
KTMAWG_ERROR_SOFTWARE_TIMEOUT = IVIC_SPECIFIC_ERROR_BASE + 0x0235
KTMAWG_ERROR_TEST_FAILED_INCORRECT_VERSION = IVIC_SPECIFIC_ERROR_BASE + 0x0236
KTMAWG_ERROR_THERMAL_SHUTDOWN = IVIC_SPECIFIC_ERROR_BASE + 0x0237
KTMAWG_ERROR_THERMAL_WARNING = IVIC_SPECIFIC_ERROR_BASE + 0x0238
KTMAWG_ERROR_UNABLE_TO_INITIALIZE_HARDWARE = IVIC_SPECIFIC_ERROR_BASE + 0x0239
KTMAWG_ERROR_UNRECOGNIZED_ERROR = IVIC_SPECIFIC_ERROR_BASE + 0x023A
KTMAWG_ERROR_UNSUPPORTED_FEATURE = IVIC_SPECIFIC_ERROR_BASE + 0x023B
KTMAWG_ERROR_UNSUPPORTED_PROPERTY = IVIC_SPECIFIC_ERROR_BASE + 0x023C
KTMAWG_ERROR_VALIDATION_ERROR = IVIC_SPECIFIC_ERROR_BASE + 0x023D
KTMAWG_ERROR_VERIFY_OPERATION_ERROR = IVIC_SPECIFIC_ERROR_BASE + 0x023E
KTMAWG_ERROR_WAIT_ABORTED = IVIC_SPECIFIC_ERROR_BASE + 0x023F
KTMAWG_ERROR_TRIGGER_NOT_SOFTWARE = IVIC_CLASS_ERROR_BASE + 0x0001
KTMAWG_ERROR_NO_WFMS_AVAILABLE = IVIC_CLASS_ERROR_BASE + 0x0004
KTMAWG_ERROR_WFM_IN_USE = IVIC_CLASS_ERROR_BASE + 0x0008
KTMAWG_ERROR_NO_SEQS_AVAILABLE = IVIC_CLASS_ERROR_BASE + 0x0009
KTMAWG_ERROR_SEQ_IN_USE = IVIC_CLASS_ERROR_BASE + 0x000D
KTMAWG_ERROR_INVALID_WFM_CHANNEL = IVIC_CLASS_ERROR_BASE + 0x000E
KTMAWG_ERROR_COMPILATION_ERROR = IVIC_SPECIFIC_ERROR_BASE + 0x0240
KTMAWG_ERROR_HARDWARE_STATE_ERROR = IVIC_SPECIFIC_ERROR_BASE + 0x0241
KTMAWG_ERROR_INCONSISTENT_MULTI_CHANNEL_PROPERTY = \
    IVIC_SPECIFIC_ERROR_BASE + 0x0242
KTMAWG_ERROR_INVALID_PARAMETER = IVIC_SPECIFIC_ERROR_BASE + 0x0243
KTMAWG_ERROR_MODULE_OUT_OF_MEMORY = IVIC_SPECIFIC_ERROR_BASE + 0x0245
KTMAWG_ERROR_SETTINGS_CONFLICT = IVIC_SPECIFIC_ERROR_BASE + 0x0246


KTMAWG_WARN_VALIDATION_FAILED = IVIC_SPECIFIC_WARN_BASE + 0x012D



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Mapping, Optional

if TYPE_CHECKING:
    import pandas as pd


class DataLengthException(Exception):
    pass


class DataPathException(Exception):
    pass


def dataframe_to_csv(
    dfdict: Mapping[str, pd.DataFrame],
    path: str,
    single_file: bool = False,
    single_file_name: Optional[str] = None,
) -> None:
    import pandas as pd

    dfs_to_save = list()
    for parametername, df in dfdict.items():
        if not single_file:
            dst = os.path.join(path, f"{parametername}.dat")
            df.to_csv(path_or_buf=dst, header=False, sep="\t")
        else:
            dfs_to_save.append(df)
    if single_file:
        df_length = len(dfs_to_save[0])
        if any(len(df) != df_length for df in dfs_to_save):
            raise DataLengthException(
                "You cannot concatenate data "
                + "with different length to a "
                + "single file."
            )
        if single_file_name is None:
            raise DataPathException(
                "Please provide the desired file name " + "for the concatenated data."
            )
        else:
            if not single_file_name.lower().endswith((".dat", ".csv", ".txt")):
                single_file_name = f"{single_file_name}.dat"
            dst = os.path.join(path, single_file_name)
            df_to_save = pd.concat(dfs_to_save, axis=1)
            df_to_save.to_csv(path_or_buf=dst, header=False, sep="\t")



import logging
from functools import partial
from typing import Any, Union

import numpy as np

import qcodes.utils.validators as vals
from qcodes.instrument.parameter import ArrayParameter, ParamRawDataType
from qcodes.instrument.visa import VisaInstrument

log = logging.getLogger(__name__)

_unit_map = {'Log mag': 'dB',
             'Phase': 'degree',
             'Delay': "",
             'Smith chart': 'dim. less',
             'Polar': 'dim. less',
             'Lin mag': 'dim. less',
             'Real': "",
             'Imaginary': "",
             'SWR': 'dim. less'}


def HPIntParser(value: str) -> int:
    """
    Small custom parser for ints

    Args:
        value: the VISA return string using exponential notation
    """
    return int(float(value))


class TraceNotReady(Exception):
    pass


class HP8753DTrace(ArrayParameter):
    """
    Class to hold a the trace from the HP8753D

    Although the trace can have two values per frequency, this
    class only returns the first value
    """

    def __init__(
            self,
            name: str,
            instrument: "HP8753D"
    ):
        super().__init__(name=name,
                         shape=(1,),  # is overwritten by prepare_trace
                         label='',  # is overwritten by prepare_trace
                         unit='',  # is overwritten by prepare_trace
                         setpoint_names=('Frequency',),
                         setpoint_labels=('Frequency',),
                         setpoint_units=('Hz',),
                         snapshot_get=False,
                         instrument=instrument
                         )

    def prepare_trace(self) -> None:
        """
        Update setpoints, units and labels
        """

        # we don't trust users to keep their fingers off the front panel,
        # so we query the instrument for all values
        assert isinstance(self.instrument, HP8753D)
        fstart = self.instrument.start_freq()
        fstop = self.instrument.stop_freq()
        npts = self.instrument.trace_points()

        sps = np.linspace(fstart, fstop, npts)
        self.setpoints = (tuple(sps),)
        self.shape = (len(sps),)

        self.label = self.instrument.s_parameter()
        self.unit = _unit_map[self.instrument.display_format()]

        self.instrument._traceready = True

    def get_raw(self) -> ParamRawDataType:
        """
        Return the trace
        """

        inst = self.instrument
        assert isinstance(inst, HP8753D)
        if not inst._traceready:
            raise TraceNotReady('Trace not ready. Please run prepare_trace.')

        inst.write('FORM2')  # 32-bit floating point numbers
        inst.write('OUTPFORM')
        inst.visa_handle.read_termination = ''
        raw_resp = inst.visa_handle.read_raw()
        inst.visa_handle.read_termination = '\n'

        first_points = b''
        # 4 bytes header, 4 bytes per point, value1's and value2's
        # are intertwined like: val1_001, val2_001, val1_002, val2_002...
        for n in range((len(raw_resp)-4)//4):
            first_points += raw_resp[4:][2*n*4:(2*n+1)*4]

        dt = np.dtype('>f')
        trace1 = np.fromstring(first_points, dtype=dt)

        return trace1


class HP8753D(VisaInstrument):
    """
    This is the QCoDeS driver for the Hewlett Packard 8753D Network Analyzer
    """

    def __init__(self, name: str, address: str, **kwargs: Any) -> None:
        super().__init__(name, address, terminator='\n', **kwargs)

        self.add_parameter(
            'start_freq',
            label='Sweep start frequency',
            unit='Hz',
            set_cmd=partial(self.invalidate_trace, 'STAR {} HZ'),
            get_cmd='STAR?',
            get_parser=float,
            vals=vals.Numbers(30000, 6000000000))

        self.add_parameter(
            'stop_freq',
            label='Sweep stop frequency',
            unit='Hz',
            set_cmd=partial(self.invalidate_trace, 'STOP {} HZ'),
            get_cmd='STOP?',
            get_parser=float,
            vals=vals.Numbers(30000, 6000000000))

        self.add_parameter(
            'averaging',
            label='Averaging state',
            set_cmd='AVERO{}',
            get_cmd='AVERO?',
            val_mapping={'ON': 1, 'OFF': 0})

        self.add_parameter(
            'number_of_averages',
            label='Number of averages',
            set_cmd='AVERFACT{}',
            get_cmd='AVERFACT?',
            get_parser=HPIntParser,
            vals=vals.Ints(0, 999))

        self.add_parameter(
            'trace_points',
            label='Number of points in trace',
            set_cmd=partial(self.invalidate_trace, 'POIN{}'),
            get_cmd='POIN?',
            get_parser=HPIntParser,
            vals=vals.Enum(3, 11, 26, 51, 101, 201, 401,
                           801, 1601))

        self.add_parameter(
            'sweep_time',
            label='Sweep time',
            set_cmd='SWET{}',
            get_cmd='SWET?',
            unit='s',
            get_parser=float,
            vals=vals.Numbers(0.01, 86400),
            )

        self.add_parameter(
            'output_power',
            label='Output power',
            unit='dBm',
            set_cmd='POWE{}',
            get_cmd='POWE?',
            get_parser=float,
            vals=vals.Numbers(-85, 20))

        self.add_parameter(
            's_parameter',
            label='S-parameter',
            set_cmd=self._s_parameter_setter,
            get_cmd=self._s_parameter_getter)

        # DISPLAY / Y SCALE PARAMETERS
        self.add_parameter(
            'display_format',
            label='Display format',
            set_cmd=self._display_format_setter,
            get_cmd=self._display_format_getter)

        # TODO: update this on startup and via display format
        self.add_parameter(
            'display_reference',
            label='Display reference level',
            unit=None,  # will be set by display_format
            get_cmd='REFV?',
            set_cmd='REFV{}',
            get_parser=float,
            vals=vals.Numbers(-500, 500))

        self.add_parameter(
            'display_scale',
            label='Display scale',
            unit=None,  # will be set by display_format
            get_cmd='SCAL?',
            set_cmd='SCAL{}',
            get_parser=float,
            vals=vals.Numbers(-500, 500))

        self.add_parameter(
            name='trace',
            parameter_class=HP8753DTrace)

        # Startup
        self.startup()
        self.connect_message()

    def reset(self) -> None:
        """
        Resets the instrument to factory default state
        """
        # use OPC to make sure we wait for operation to finish
        self.ask('OPC?;PRES')

    def run_continously(self) -> None:
        """
        Set the instrument in run continously mode
        """
        self.write('CONT')

    def run_N_times(self, N: int) -> None:
        """
        Run N sweeps and then hold. We wait for a response before returning
        """

        st = self.sweep_time.get_latest()

        if N not in range(1, 1000):
            raise ValueError(f'Can not run {N} times.' +
                             ' please select a number from 1-999.')

        # set a longer timeout, to not timeout during the sweep
        new_timeout = st*N + 2

        with self.timeout.set_to(new_timeout):
            log.debug(f'Making {N} blocking sweeps.' +
                      f' Setting VISA timeout to {new_timeout} s.')

            self.ask(f'OPC?;NUMG{N}')

    def invalidate_trace(self, cmd: str,
                         value: Union[float, int, str]) -> None:
        """
        Wrapper for set_cmds that make the trace not ready
        """
        self._traceready = False
        self.write(cmd.format(value))

    def startup(self) -> None:
        self._traceready = False
        self.display_format(self.display_format())

    def _s_parameter_setter(self, param: str) -> None:
        """
        set_cmd for the s_parameter parameter
        """
        if param not in ['S11', 'S12', 'S21', 'S22']:
            raise ValueError('Cannot set s-parameter to {}')

        # the trace labels changes
        self._traceready = False

        self.write(param)

    def _s_parameter_getter(self) -> str:
        """
        get_cmd for the s_parameter parameter
        """
        for cmd in ['S11?', 'S12?', 'S21?', 'S22?']:
            resp = self.ask(cmd)
            if resp in ['1', '1\n']:
                break

        return cmd.replace('?', '')

    def _display_format_setter(self, fmt: str) -> None:
        """
        set_cmd for the display_format parameter
        """
        val_mapping = {'Log mag': 'LOGM',
                       'Phase': 'PHAS',
                       'Delay': 'DELA',
                       'Smith chart': 'SMIC',
                       'Polar': 'POLA',
                       'Lin mag': 'LINM',
                       'Real': 'REAL',
                       'Imaginary': 'IMAG',
                       'SWR': 'SWR'}

        if fmt not in val_mapping.keys():
            raise ValueError(f'Cannot set display_format to {fmt}.')

        self._traceready = False
        self.display_reference.unit = _unit_map[fmt]
        self.display_scale.unit = _unit_map[fmt]

        self.write(val_mapping[fmt])

    def _display_format_getter(self) -> str:
        """
        get_cmd for the display_format parameter
        """
        val_mapping = {'LOGM': 'Log mag',
                       'PHAS': 'Phase',
                       'DELA': 'Delay',
                       'SMIC': 'Smith chart',
                       'POLA': 'Polar',
                       'LINM': 'Lin mag',
                       'REAL': 'Real',
                       'IMAG': 'Imaginary',
                       'SWR': 'SWR'}

        # keep asking until we find the currently used format
        for cmd in val_mapping.keys():
            resp = self.ask(f'{cmd}?')
            if resp in ['1', '1\n']:
                break

        return val_mapping[cmd]



# pylint: disable=redefined-outer-name
import pytest
from hypothesis import given
import hypothesis.strategies as st
from qcodes.instrument_drivers.Keysight.keysight_34980a import Keysight34980A
from qcodes.instrument_drivers.Keysight.keysight_34934a import Keysight34934A
import qcodes.instrument.sims as sims

VISALIB = sims.__file__.replace('__init__.py', 'keysight_34980A.yaml@sim')


@pytest.fixture(scope="module")
def switch_driver():
    inst = Keysight34980A('keysight_34980A_sim',
                          address='GPIB::1::INSTR',
                          visalib=VISALIB)

    try:
        yield inst
    finally:
        inst.close()


def test_protection_mode(switch_driver):
    """
    to check the protection mode (34934A module only)
    """
    assert switch_driver.module[1].protection_mode() == 'AUTO100'


def test_connection(switch_driver):
    """
    to check if a channel is closed or open
    """
    assert not switch_driver.module[1].is_closed(2, 3)
    assert switch_driver.module[1].is_open(2, 3)

# The following is to test the numbering function for the module 34934A
# the 'g' functions are copied from the table on P168 of the 34934A User's Guide
# the 'f' function is a simplified version, see the keysight34934A class for
# detail

@given(
    st.sampled_from(("M1H", "M1L", "M2H", "M2L")),
    st.integers(1, 4),
    st.integers(1, 32)
)
def test_4x32(config, row, column):
    f = Keysight34934A.get_numbering_function(4, 32, config)
    g = numbering_function_4x32(config)
    assert f(row, column) == g(row, column)


@given(
    st.sampled_from(("MH", "ML")),
    st.integers(1, 4),
    st.integers(1, 64)
)
def test_4x64(config, row, column):
    f = Keysight34934A.get_numbering_function(4, 64, config)
    g = numbering_function_4x64(config)
    assert f(row, column) == g(row, column)


@given(
    st.integers(1, 4),
    st.integers(1, 128)
)
def test_4x128(row, column):
    f = Keysight34934A.get_numbering_function(4, 128)
    g = numbering_function_4x128()
    assert f(row, column) == g(row, column)


@given(
    st.sampled_from(("MH", "ML")),
    st.integers(1, 8),
    st.integers(1, 32)
)
def test_8x32(config, row, column):
    f = Keysight34934A.get_numbering_function(8, 32, config)
    g = numbering_function_8x32(config)
    assert f(row, column) == g(row, column)


@given(
    st.integers(1, 8),
    st.integers(1, 64)
)
def test_8x64(row, column):
    f = Keysight34934A.get_numbering_function(8, 64)
    g = numbering_function_8x64()
    assert f(row, column) == g(row, column)


@given(
    st.integers(1, 16),
    st.integers(1, 32)
)
def test_16x32(row, column):
    f = Keysight34934A.get_numbering_function(16, 32)
    g = numbering_function_16x32()
    assert f(row, column) == g(row, column)


def numbering_function_4x32(wiring_config):
    offsets = {
        "M1H": 0,
        "M2H": 32,
        "M1L": 64,
        "M2L": 96
    }

    def numbering_function(row, col):
        n = 100 * (2 * row - 1) + col + offsets[wiring_config]
        return str(int(n))

    return numbering_function


def numbering_function_4x64(wiring_config):
    offsets = {
        "MH": 0,
        "ML": 64
    }

    def numbering_function(row, col):
        n = 100 * (2 * row - 1) + col + offsets[wiring_config]
        return str(int(n))

    return numbering_function


def numbering_function_4x128():
    def numbering_function(row, col):
        n = 100 * (2 * row - 1) + col
        return str(int(n))

    return numbering_function


def numbering_function_8x32(wiring_config):
    offsets = {
        "MH": 0,
        "ML": 32
    }

    def numbering_function(row, col):
        n = 100 * row + col + offsets[wiring_config]
        return str(int(n))

    return numbering_function


def numbering_function_8x64():
    def numbering_function(row, col):
        n = 100 * row + col
        return str(int(n))

    return numbering_function


def numbering_function_16x32():
    def numbering_function(row, col):
        n = 50 * (row + 1) + col
        return str(int(n))

    return numbering_function
